﻿/// @file parsewav.cpp
///
/// @brief waveファイルを解析
///
#include "parsewav.h"
#include <wx/filename.h>
#include "utils.h"
#include "version.h"

//#define MAX_CARRIER_PATTERN	1
static const char	*carrier_pattern[4][4]={
	{
		"11001100",	// 0 (600 baud)
		"00110011",	// 0 (600 baud frip)
		"10101010",	// 1 (600 baud)
		"01010101",	// 1 (600 baud frip)
	},{
		"1100",		// 0 (1200 baud)
		"0011",		// 0 (1200 baud frip)
		"1010",		// 1 (1200 baud)
		"0101",		// 1 (1200 baud frip)
	},{
		"11",		// 0 (2400 baud)
		"00",		// 0 (2400 baud frip)
		"10",		// 1 (2400 baud)
		"01",		// 1 (2400 baud frip)
	},{
		"1100110011001100",	// 0 (300 baud)
		"0011001100110011",	// 0 (300 baud frip)
		"1010101010101010",	// 1 (300 baud)
		"0101010101010101",	// 1 (300 baud frip)
	}
};
#define IDX_PTN_600		0
#define IDX_PTN_1200	1
#define IDX_PTN_2400	2
#define IDX_PTN_300		3
// 2400ボーの時はエッジ判定を優先
static const char *carrier_edge_pattern[]={
	"0010",		// 01 (2400 baud)
	"1101",		// 01 (2400 baud frip)
	"0100",		// 10 (2400 baud)
	"1011",		// 10 (2400 baud frip)
	NULL
};

// ボーレート自動解析用
// start_bit + data_bit($ff) + 2stop_bit
static const struct st_conv_baud_tbl_ff {
	int   baud;
	int   len;
	const char *ptn;
} conv_baud_tbl_ff[]={
	{  300, 88, "0000000011111111111111111111111111111111111111111111111111111111111111111111111111111111" },
	{  600, 44, "00001111111111111111111111111111111111111111" },
	{ 1200, 22, "0011111111111111111111" },
	{ 2400, 11, "01111111111" },
};
static const int c_baud_min[4]={ 3, 0, 1, 2 };

static const int c_baud_ridx[4]={ 1, 2, 3, 0 };

static const double c_baud_rate[4] = { 600.0, 1200.0, 2400.0, 300.0 };

static const uint8_t cmt_ident[4] = { 0xff, 0x01, 0x3c, 0x00 };

static const int c_gap_length[4] = { 0x5a, 0xc0, 0x19b, 0x36f };

#define T9X_IDENTIFIER "eMB-689X CassetteTapeImageFile  "

/// @brief コンストラクタ
///
PARSEWAV::PARSEWAV(wxWindow *parent)
{
	parent_window = parent;

	//
	wave_data = new Buf_t<uint8_t>(true);
	wave_correct_data = new Buf_t<uint8_t>(true);
	carrier_data = new Buf_t<char>(true);
	serial_data = new Buf_t<char>(true);
	serial_new_data = new Buf_t<char>(true, true);
	binary_data = new Buf_t<uint8_t>(true, true);

//	buffer_size = DATA_BUFFER_SIZE - 2;

	infile_type = UNKNOWN;
	outfile_type = UNKNOWN;

	fio_in  = NULL;
	fio_out = NULL;
	fio_outs = NULL;
	fiolog  = NULL;

	logbuf = NULL;
	logfile = _T("wavtool.log");

	include_header = true;

	err_num = pwErrNone;

	chk_wav_mode = false;

	progbox = new ProgressBox(parent_window, wxID_ANY);
	rftypebox = new RfTypeBox(parent_window, wxID_ANY);
	maddressbox = new MAddressBox(parent_window, wxID_ANY);
	errinfo = new PwErrInfo();

	// wavファイルヘッダ(出力用)
	outwav.Init(48000, 8, 1);
}

/// @brief デストラクタ
///
PARSEWAV::~PARSEWAV()
{
	CloseDataFile();
	CloseOutFile();

	std::vector<REPORT4 *>::iterator itm = rep4.begin();
	while ( itm != rep4.end() ) {
		delete *itm;
		itm++;
	}
	rep4.clear();

	delete errinfo;
	delete maddressbox;
	delete rftypebox;
	delete progbox;

	delete binary_data;
	delete serial_new_data;
	delete serial_data;
	delete carrier_data;
	delete wave_correct_data;
	delete wave_data;
}


/// @brief 入力ファイルのフォーマットをチェックする
///
/// @param[in]     fp        入力ファイル
/// @param[in]     in_file   ファイルパス名
/// @param[in,out] file_type ファイル種類
/// @return    pwOK:正常 pwError:エラーあり pwWarn:警告あり
PwErrType PARSEWAV::CheckFileFormat(FILE *fp, const _TCHAR *in_file, enum_file_type *file_type) {
	PwErrType rc;

	// check file size
	char buf[64];
	size_t len = fread(buf, sizeof(char), sizeof(buf), fp);
	if (len == 0) {
		err_num = pwErrFileEmpty;
		errinfo->SetInfo(__LINE__, pwError, err_num);
		errinfo->ShowMsgBox();
		return pwError;
	}
	fseek(fp, 0, SEEK_SET);

	// check wave header
	if (*file_type == WAV) {
		rc = check_wav_format(fp, inwav.GetHead(), inwav.GetFmtChank(), inwav.GetDataChank());
	} else if (*file_type == L3C) {
		rc = check_l3c_format(fp);
	} else if (*file_type == L3B) {
		rc = check_l3b_format(fp);
	} else if (*file_type == T9X) {
		rc = check_t9x_format(fp);
	} else if (*file_type == L3) {
		rc = check_l3_format(fp);
	} else {
		rc = check_rf_format(fp, in_file, file_type);
	}

	return rc;
}

/// @brief 入力ファイルのフォーマットからデータ開始位置をシーク
///
/// @param[in] fp        入力ファイル
/// @param[in] in_file   ファイルパス名
/// @param[in] file_type ファイル種類
/// @return    pwOK:正常 pwError:エラーあり pwWarn:警告あり
PwErrType PARSEWAV::SeekFileFormat(FILE *fp, const _TCHAR *in_file, enum_file_type file_type) {
	PwErrType rc = pwOK;

	// check file size
	char buf[64];
	size_t len = fread(buf, sizeof(char), sizeof(buf), fp);
	if (len == 0) {
		err_num = pwErrFileEmpty;
		errinfo->SetInfo(__LINE__, pwError, err_num);
		errinfo->ShowMsgBox();
		return pwError;
	}
	fseek(fp, 0, SEEK_SET);

	// seek header
	if (file_type == WAV) {
		rc = check_wav_format(fp, inwav.GetHead(), inwav.GetFmtChank(), inwav.GetDataChank());
	} else if (file_type == T9X) {
		rc = check_t9x_format(fp);
	}

	return rc;
}

///
///
///
///
///
void PARSEWAV::init_parameters()
{
	st_pa_carr.x0_prev = 0.0;
	st_pa_carr.sample_cnt = 0;
	st_pa_carr.wav_prev = 0;
	st_pa_carr.carr_prev = 0;
	st_pa_carr.odd = 0;

	st_lamda.us_delta = 1000000.0 / (double)inwav.GetSampleRate();
	for(int n=0; n<3; n++) {
		st_lamda.samples[n] = (double)inwav.GetSampleRate() / (double)(1200 << n);

		st_lamda.us_range[n] = 1000000.0 / (double)(1200 << n) / (double)(1 << half_wave);
		st_lamda.us[n] = 1000000.0 / (double)param.GetFreq(n) / (double)(1 << half_wave);
		st_lamda.us_avg[n] = st_lamda.us[n];
	}
	for(int n=0; n<2; n++) {
		st_lamda.us_mid[n] = 1000000.0 / (double)(1800 << n) / (double)(1 << half_wave);
		st_lamda.us_mid_avg[n] = st_lamda.us_mid[n];

		st_lamda.us_limit[n] = 1000000.0 / (double)(1200 << (n + 3));
	}

	int ns = param.GetFskSpeed();
	if (chk_wav_mode == true) {
		for(int n=0; n<2; n++, ns++) {
			st_lamda.us_min[ns] = st_lamda.us[ns] - st_lamda.us_range[ns] * 25.0 / 100.0;
			st_lamda.us_max[ns] = st_lamda.us[ns] + st_lamda.us_range[ns] * 50.0 / 100.0;
		}
	} else {
		for(int n=0; n<2; n++, ns++) {
			st_lamda.us_min[ns] = st_lamda.us[ns] - st_lamda.us_range[ns] * param.GetRange(n) / 100.0;
			st_lamda.us_max[ns] = st_lamda.us[ns] + st_lamda.us_range[ns] * param.GetRange(n) / 100.0;
		}
	}
}

///
///
///
///
///
void PARSEWAV::SetFrequency(int magnify)
{
	param.SetFrequency(magnify);
}

/// @brief ファイルのヘッダをファイルに出力する
///
/// @param[in,out] fp        出力ファイル
/// @param[in]     file_type ファイル種類
/// @return ヘッダ長さ
///
int PARSEWAV::InitFileHeader(FILE *fp, enum_file_type file_type)
{
	int len_all = 0;

	if (fp == NULL) return len_all;

	if (file_type == WAV) {
		// wavファイルのサンプルレート、ビットなどを設定
		outwav.Init(get_sample_rate(param.GetSampleRate()), (param.GetSampleBits() == 0 ? 8 : 16), 1);

		conv.InitConvSampleData();

		len_all = outwav.Out(fp);

	} else if (file_type == T9X) {
		t9x_header_t head;

		memset(&head, 0, sizeof(head));
		memcpy(head.ident, T9X_IDENTIFIER, sizeof(head.ident));
		fwrite((void *)&head, sizeof(head), 1, fp);
		len_all = sizeof(head);
	}

	return len_all;
}

/// @brief ファイルにファイルサイズを書き込む
///
/// @param[in,out] fp            出力ファイル
/// @param[in]     in_file_type  入力ファイル種類
/// @param[in]     out_file_type 出力ファイル種類
///
///
void PARSEWAV::SetFileHeader(FILE *fp, enum_file_type in_file_type, enum_file_type out_file_type)
{
	if (fp == NULL) return;

	if (out_file_type == WAV) {
		// WAVファイル以外からの変換の場合、データ末尾に2400/4800Hzを0.5sec出力
		if (in_file_type != WAV) {
			int spd = param.GetFskSpeed();
			int samples = 48000 / 2400 / (spd + 1);
			uint8_t w_data[200];
			for(int j=0; j<120; j++) {
				char c = carrier_data->get(-1);	// 最後のデータ
				carrier_data->shift();
				for(int i=0; i<samples; i++) {
					c = ((1 - (c & 1)) | 0x30);
					carrier_data->add(c);
				}
				int pos = 0;
				while(carrier_data->r_pos < carrier_data->w_pos) {
					pos += encode_to_wave_data(carrier_data, &w_data[pos]);
					carrier_data->r_pos++;
					carrier_data->total_r_pos++;
				}
				// convert
				conv.OutConvSampleData(w_data,48000,1,pos,outwav.GetSampleRate(),outwav.GetBlockSize(),fp);
			}
		}
		// ファイルサイズ、データサイズを書き込む
		PARSEWAV_FORMAT::OutSize(fp);

	} else if (out_file_type == T9X) {
		t9x_header_t head;

		memset(&head, 0, sizeof(head));
		memcpy(head.ident, T9X_IDENTIFIER, sizeof(head.ident));
		head.data2 = 9;	// TODO: what's this?

		fseek(fp, 0, SEEK_SET);
		fwrite((void *)&head, sizeof(head), 1, fp);

	}

	return;
}

/// @brief サンプルレートを返す
///
/// @param[in] num 0:11025Hz 1:22050Hz 2:44100Hz
/// @return サンプルレート
int PARSEWAV::get_sample_rate(int num) {
	int rate = 48000;

	switch(num) {
		case 0:
			rate = 11025;
			break;
		case 1:
			rate = 22050;
			break;
		case 2:
			rate = 44100;
			break;
	}
	return rate;
}


/// @brief wavファイルのフォーマットをチェック
///
/// @param[in] fp 入力ファイル
/// @return pwOK
///
PwErrType PARSEWAV::check_wav_format(FILE *fp, wav_header_t *head, wav_fmt_chank_t *fmt, wav_data_chank_t *data)
{
	err_num = conv.CheckWavFormat(fp, head, fmt, data);
	if (err_num != pwErrNone) {
		errinfo->SetInfo(__LINE__, pwError, err_num);
		errinfo->ShowMsgBox();
		return pwError;
	}

	sample_num = (data->data_len) / fmt->channels;
	if(fmt->sample_bits == 16) {
		sample_num /= 2;
	}
	sample_num_orig = sample_num;
	sample_pos = 0;
	sample_ms = 0;

	return pwOK;
}

/// @brief l3cファイルのフォーマットをチェック（チェックしていないが）
///
/// @param[in] fp 入力ファイル
/// @return pwOK
///
PwErrType PARSEWAV::check_l3c_format(FILE *fp)
{
	fseek(fp, 0, SEEK_END);
	sample_num = ftell(fp);
	fseek(fp, 0, SEEK_SET);

	sample_num_orig = sample_num;
	sample_pos = 0;
	sample_ms = 0;

	return pwOK;
}

/// @brief l3bファイルのフォーマットをチェック（チェックしていないが）
///
/// @param[in] fp 入力ファイル
/// @return pwOK
///
PwErrType PARSEWAV::check_l3b_format(FILE *fp)
{
	fseek(fp, 0, SEEK_END);
	sample_num = ftell(fp);
	fseek(fp, 0, SEEK_SET);

	sample_num_orig = sample_num;
	sample_pos = 0;
	sample_ms = 0;

	return pwOK;
}

/// @brief t9xファイルのフォーマットをチェック
///
/// @param[in] fp 入力ファイル
/// @return pwOK
///
PwErrType PARSEWAV::check_t9x_format(FILE *fp)
{
	t9x_header_t head;

	fread(&head, sizeof(t9x_header_t), 1, fp);
	if(memcmp(head.ident,T9X_IDENTIFIER,sizeof(head.ident)) != 0) {
		// this is not t9x format
		err_num = pwErrNotT9XFormat;
		errinfo->SetInfo(__LINE__, pwError, err_num);
		errinfo->ShowMsgBox();
		return pwError;
	}

	fseek(fp, 0, SEEK_END);
	sample_num = ftell(fp);

	// 実際のサンプルデータがはじまる位置の先頭
	fseek(fp, sizeof(head), SEEK_SET);

	sample_num -= sizeof(head);	// 64bytes

	sample_num_orig = sample_num;
	sample_pos = 0;
	sample_ms = 0;

	return pwOK;
}

/// @brief l3ファイルのフォーマットをチェック（チェックしていないが）
///
/// @param[in] fp 入力ファイル
/// @return pwOK
///
PwErrType PARSEWAV::check_l3_format(FILE *fp)
{
	fseek(fp, 0, SEEK_END);
	sample_num = ftell(fp);
	fseek(fp, 0, SEEK_SET);

	sample_num_orig = sample_num;
	sample_pos = 0;
	sample_ms = 0;

	return pwOK;
}

/// @brief 実ファイルのフォーマットをチェック
///
/// @param[in] fp 入力ファイル
/// @param[in] in_file 入力ファイル名
/// @param[in,out] file_type 入力ファイル種類
/// @return pwOK:OK pwCancel:キャンセル終了
PwErrType PARSEWAV::check_rf_format(FILE *fp, const _TCHAR *in_file, enum_file_type *file_type)
{
	int rc;
	_TCHAR bname[_MAX_PATH];

	UTILS::base_name(in_file, bname);

	// ファイルの種類を選択
	rc = rftypebox->showRftypeBox(bname, true);
	if (rc != 1) {
		// cancel button
		return pwCancel;
	}

	PwErrType rt = get_first_rf_data(fp, file_type);

	return rt;
}

/// @brief 実ファイルを読みそのファイル種類を判定する
///
/// @param[in] fp 入力ファイル
/// @return pwOK:OK pwWarning:警告あり
///
PwErrType PARSEWAV::get_first_rf_data(FILE *fp, enum_file_type *file_type)
{
	PwErrType rc = pwOK;
	int dfmt;
	int dtype;
	int addr;

	include_header = true;

	fseek(fp, 0, SEEK_END);
	sample_num = ftell(fp);
	fseek(fp, 0, SEEK_SET);

	sample_num_orig = sample_num;
	sample_pos = 0;
	sample_ms = 0;

	switch(rftypebox->getRfDataFileType()) {
	case 0:
		// ただのファイル指定の場合
		*file_type = PARSEWAV::PLAIN;
		break;
	default:
		// 実ファイル指定の場合
		*file_type = PARSEWAV::REAL;
		dfmt = rftypebox->getRfDataFormat();
		dtype = rftypebox->getRfDataType();
		switch(dfmt) {
			case 0:	// basic
				if (dtype != 0xff) { // 中間言語形式の場合
					// ファイルの最初を読み込む
					fread(rf_header, sizeof(uint8_t), sizeof(rf_header), fp);
					if (rf_header[0] != 0xff && rf_header[0] != 0xfe) {
						// maybe not have header
						err_num = pwErrNoBASICIntermediateLanguage;
						rc = pwWarning;
						errinfo->SetInfo(__LINE__, rc, err_num);
						errinfo->ShowMsgBox();
					}
				}
				break;

			case 2: // machine

				// ファイルの最初/最後を読み込む
				fread(rf_header, sizeof(uint8_t), sizeof(rf_header), fp);
				fseek(fp, -5, SEEK_END);
				fread(rf_footer, sizeof(uint8_t), sizeof(rf_footer), fp);

				if (rf_header[0] != 0x00 || rf_footer[0] != 0xff || rf_footer[1] != 0x00) {
					// maybe not have header and footer
					include_header = false;
					maddressbox->setValid(false);
					maddressbox->setDataSize(sample_num);
					maddressbox->setFileSize(sample_num);
				} else {
					maddressbox->setValid(true);
					addr = rf_header[3] * 256 + rf_header[4];
					maddressbox->setStartAddr(addr);
					addr = rf_header[1] * 256 + rf_header[2];
					maddressbox->setDataSize(addr);
					addr = rf_footer[3] * 256 + rf_footer[4];
					maddressbox->setExecAddr(addr);
					maddressbox->setIncHeader(include_header);
					maddressbox->setFileSize(sample_num);
				}
				// ダイアログ表示
				ShowMAddressBox(true);
				break;
		}
		break;
	}
	fseek(fp, 0, SEEK_SET);

	return rc;
}


/// @brief wavファイルから１サンプル読む
///
/// @param[in] size    1: uint8_tで返す  2: int16_tで返す
/// @param[in] reverse 1: 波形を反転
/// @return 変換した１サンプルデータ
///
int PARSEWAV::get_wav_sample(int size, int reverse)
{
	int l = 0;
	int h = 0;
	int in_bits = inwav.GetSampleBits();

	// 1サンプルを得る
	l = fgetc(fio_in);
	if(in_bits == 16) {
		// int16_tの場合
		h = fgetc(fio_in);
		if (h >= 128) {
			h = h - 256;
		}
		if (reverse) {
			h *= -1;
			if (h >= 128) h = 127;
			l *= -1;
		}
	} else {
		if (reverse) {
			l = 256 - l;
			if (l >= 256) l = 255;
		}
	}
	// ステレオの場合、左側を無視する
	for(int i = 2; i <= inwav.GetChannels(); i++) {
		fgetc(fio_in);
		if(in_bits == 16) {
			fgetc(fio_in);
		}
	}
	sample_pos++;
	sample_ms = (uint32_t)(1000.0 * (double)sample_pos / (double)inwav.GetSampleRate());

	if (size == 1) {
		if(in_bits == 16) {
			// int16_t -> uint8_t
			return ((h + 128) & 0xff);
		} else {
			return (l & 0xff);
		}
	} else {
		if(in_bits == 16) {
			return h * 256 + l;
		} else {
			// uint8_t -> int16_t
			return (l - 128) * 256;
		}
	}
}
/// @brief wavファイルから１サンプル読む(int16で返す)
///
/// @param[in] reverse 1: 波形を反転
/// @return 変換した１サンプルデータ
int PARSEWAV::get_wav_sample(int reverse)
{
	return get_wav_sample(2, reverse);
}

/// @brief wavファイルから１サンプルを読んでバッファに追記
///
/// @param[in,out] w_data サンプルデータ用のバッファ(追記していく)
/// @param[in] reverse 1: 波形を反転
/// @return 読み込んだデータの長さ
///
int PARSEWAV::get_wav_sample(Buf_t<uint8_t> *w_data, int reverse)
{
	int l;

	while(w_data->isFull(0) != true && sample_pos < sample_num) {
		l = get_wav_sample(1, reverse);

		w_data->add((uint8_t)l, sample_ms);
	}
	if (sample_pos >= sample_num) {
		w_data->last = true;
	}
	return w_data->w_pos;
}

/// @brief l3cファイル(搬送波ビットデータ)から１データ読んでバッファに追記
///
/// @param[in,out] c_data 搬送波データ用のバッファ(追記していく)
/// @return 読み込んだデータの長さ
///
///
int PARSEWAV::get_l3c_sample(Buf_t<char> *c_data)
{
//	int pos = 0;
	int l;

//	pos = strlen(data);
	while(c_data->isFull(0) != true && sample_pos < sample_num) {
		l = fgetc(fio_in);
		l &= 0xff;

		sample_pos++;
		sample_ms = (uint32_t)(1000.0 * (double)sample_pos / 4800.0);
		if (l == '\r' || l == '\n') {
			continue;
		}
		c_data->add(l, sample_ms);
	}
	if (sample_pos >= sample_num) {
		c_data->last = true;
	}
	return c_data->w_pos;
}

/// @brief l3bファイルから１データ読んでバッファに追記
///
/// @param[in,out] s_data シリアルデータ用のバッファ(追記していく)
/// @return 読み込んだデータの長さ
///
///
int PARSEWAV::get_l3b_sample(Buf_t<char> *s_data)
{
//	int pos = 0;
	int l;
	int baud_mag = param.GetFskSpeed() + 1;

//	pos = strlen(data);
	while(s_data->isFull(0) != true && sample_pos < sample_num) {
		l = fgetc(fio_in);
		l &= 0xff;

		sample_pos++;
		sample_ms = (uint32_t)(1000.0 * (double)sample_pos / (c_baud_rate[param.GetBaud()] * (double)baud_mag));
		if (l == '\r' || l == '\n') {
			continue;
		}
		s_data->add(l, sample_ms);
	}
	if (sample_pos >= sample_num) {
		s_data->last = true;
	}
	return s_data->w_pos;
}

/// @brief t9xファイルから１データ読んでバッファに追記
///
/// @param[in,out] s_data シリアルデータ用のバッファ(追記していく)
/// @return 読み込んだデータの長さ
///
///
int PARSEWAV::get_t9x_sample(Buf_t<char> *s_data)
{
	int l;
	int baud_mag = param.GetFskSpeed() + 1;

	while(s_data->isFull(0) != true && sample_pos < sample_num) {
		l = fgetc(fio_in);
		l &= 0xff;

		sample_pos++;
		for(int i=0; i<8; i++) {
			char c = (l & (1 << i)) ? '1' : '0';
			sample_ms = (uint32_t)(1000.0 * (double)(sample_pos * 8 + i) / (c_baud_rate[param.GetBaud()] * (double)baud_mag));
			s_data->add(c, sample_ms);
		}
	}
	if (sample_pos >= sample_num) {
		s_data->last = true;
	}
	return s_data->w_pos;
}

/// @brief l3ファイルからデータ読んでバッファに追記
///
/// @param[in,out] b_data バイナリデータ用のバッファ(追記していく)
/// @return 読み込んだデータの長さ
///
int PARSEWAV::get_l3_sample(Buf_t<uint8_t> *b_data)
{
	int pos = 0;

	pos = (int)fread(&(b_data->data[b_data->w_pos]), sizeof(uint8_t), b_data->size-(b_data->w_pos), fio_in);
	sample_pos += pos;
	(b_data->w_pos) += pos;

	if (sample_pos >= sample_num) {
		b_data->last = true;
	}
	return (b_data->w_pos);
}

/// @brief l3ファイルから１データ読む
///
/// @return １バイトデータ
///
///
uint8_t PARSEWAV::get_l3_sample()
{
	int l;

	l = fgetc(fio_in);

	sample_pos++;

	return (l & 0xff);
}

/// @brief 実ファイルから１データ読む
///
/// @return １バイトデータ
///
///
uint8_t PARSEWAV::get_rf_sample()
{
	int l;

	if (save_data_name[8] == 2) {
		// マシン語
		if (sample_pos < 5) {
			if (include_header) l = fgetc(fio_in);
			l = rf_header[sample_pos];
		} else if (sample_num - sample_pos <= 5) {
			if (include_header) l = fgetc(fio_in);
			l = rf_footer[5 - sample_num + sample_pos];
		} else {
			l = fgetc(fio_in);
		}
	} else {
		l = fgetc(fio_in);
	}
	sample_pos++;

	return (l & 0xff);
}

/// @brief 2400Hzと1200Hzの波を見つけてビットデータに変換
///
/// @param[in]   fsk_spd   1:倍速FSK
/// @param[in]   w_data    WAVサンプルデータ
/// @param[out]  c_data  "10": 2400Hz "1100":1200Hz "?HHLL":Too long
/// @return WAV1サンプルデータ
int PARSEWAV::decode_to_carrier_wave_bn(int fsk_spd, Buf_t<uint8_t> *w_data, Buf_t<char> *c_data)
{
	char bit_data[10];

	int data = 0;
	uint8_t  dat;
	uint32_t ms = 0;
	bool hi = false;
	bool found = false;

	double x0 = 0.0;
	double lamda = 0.0;

	// find next trigger point
	while(w_data->isEOB(0) != true) {
		w_data->get(dat, ms);
		data = (int)dat - 128;

		if (st_pa_carr.wav_prev < 0 && data >= 0) {
			// l -> h になる位置から波長(us)を求める
			hi = true;
			x0 = - (st_pa_carr.wav_prev * st_lamda.us_delta / (data - st_pa_carr.wav_prev));
			lamda = st_pa_carr.x0_prev + (st_pa_carr.sample_cnt * st_lamda.us_delta) + x0;
			// 波長が短すぎる場合はノイズと判断して処理継続
			if (lamda >= st_lamda.us_limit[fsk_spd]) {
				st_pa_carr.x0_prev = st_lamda.us_delta - x0;
				st_pa_carr.odd = 1 - st_pa_carr.odd;
				found = true;
				break;
			}
		} else if (st_pa_carr.wav_prev > 0 && data <= 0 && half_wave == 1) {
			// h -> l になる位置から波長(us)を求める
			hi = false;
			x0 = - (st_pa_carr.wav_prev * st_lamda.us_delta / (data - st_pa_carr.wav_prev));
			lamda = st_pa_carr.x0_prev + (st_pa_carr.sample_cnt * st_lamda.us_delta) + x0;
			// 波長が短すぎる場合はノイズと判断して処理継続
			if (lamda >= st_lamda.us_limit[fsk_spd]) {
				st_pa_carr.x0_prev = st_lamda.us_delta - x0;
				st_pa_carr.odd = 1 - st_pa_carr.odd;
				found = true;
				break;
			}
		}
		st_pa_carr.sample_cnt++;
		st_pa_carr.wav_prev = data;
	}
	if (found != true) {
		return data;
	}

	if (half_wave == 0) {
		// 全波
		if (st_lamda.us_min[fsk_spd] <= lamda && lamda <= st_lamda.us_max[fsk_spd]) {
			// long (half wave)
			strcpy(bit_data,"1100");
			st_pa_carr.carr_prev = 4;
			st_lamda.us_avg[fsk_spd] = (st_lamda.us_avg[fsk_spd] + lamda) / 2.0;

			rep_p1_sample_num[0]++;
		}
		else if (st_lamda.us_min[fsk_spd+1] <= lamda && lamda <= st_lamda.us_max[fsk_spd+1]) {
			// short (half wave)
			strcpy(bit_data,"10");
			st_pa_carr.carr_prev = 2;
			st_lamda.us_avg[fsk_spd+1] = (st_lamda.us_avg[fsk_spd+1] + lamda) / 2.0;

			rep_p1_sample_num[1]++;
		}
		else if (st_lamda.us_mid[fsk_spd] <= lamda && lamda < st_lamda.us_min[fsk_spd]) {
			// middle long
			strcpy(bit_data,"?1100");
			st_pa_carr.carr_prev = 4;
			st_lamda.us_mid_avg[fsk_spd] = (st_lamda.us_mid_avg[fsk_spd] + lamda) / 2.0;

			rep_p1_sample_num[2]++;
		}
		else if (st_lamda.us_max[fsk_spd+1] < lamda && lamda < st_lamda.us_mid[fsk_spd]) {
			// middle short
			strcpy(bit_data,"?10");
			st_pa_carr.carr_prev = 2;
			st_lamda.us_mid_avg[fsk_spd] = (st_lamda.us_mid_avg[fsk_spd] + lamda) / 2.0;

			rep_p1_sample_num[2]++;
		}
		else if (st_lamda.us_max[fsk_spd] < lamda) {
			// too long
			strcpy(bit_data,"?HHLL");
			st_pa_carr.carr_prev = 5;
			rep_p1_sample_num[3]++;
		}
		else if (lamda < st_lamda.us_min[fsk_spd+1]) {
			// too short
			strcpy(bit_data,"?HL");
			st_pa_carr.carr_prev = 1;
			rep_p1_sample_num[4]++;
		}
		else {
			// error
			strcpy(bit_data,"??");
			st_pa_carr.carr_prev = 0;
			rep_p1_sample_num[5]++;
		}
	} else {
		// 半波
		if (st_lamda.us_min[fsk_spd] <= lamda && lamda <= st_lamda.us_max[fsk_spd]) {
			// long (half wave)
			if (hi) {
				strcpy(bit_data,"00");
				if (st_pa_carr.carr_prev != 4) st_pa_carr.odd = 1;
				st_pa_carr.carr_prev = -4;
			} else {
				strcpy(bit_data,"11");
				if (st_pa_carr.carr_prev != -4) st_pa_carr.odd = 1;
				st_pa_carr.carr_prev = 4;
			}
			st_lamda.us_avg[fsk_spd] = (st_lamda.us_avg[fsk_spd] + lamda) / 2.0;

			rep_p1_sample_num[0]++;
		}
		else if (st_lamda.us_min[fsk_spd+1] <= lamda && lamda <= st_lamda.us_max[fsk_spd+1]) {
			// short (half wave)
			if (hi) {
				strcpy(bit_data,"0");
				if (st_pa_carr.carr_prev != 2) st_pa_carr.odd = 1;
				st_pa_carr.carr_prev = -2;
			} else {
				strcpy(bit_data,"1");
				if (st_pa_carr.carr_prev != -2) st_pa_carr.odd = 1;
				st_pa_carr.carr_prev = 2;
			}
			st_lamda.us_avg[fsk_spd+1] = (st_lamda.us_avg[fsk_spd+1] + lamda) / 2.0;

			rep_p1_sample_num[1]++;
		}
		else if (st_lamda.us_max[fsk_spd+1] < lamda && lamda < st_lamda.us_min[fsk_spd]) {
			// middle
			if (hi) {
				if ((st_pa_carr.odd == 0 && st_pa_carr.carr_prev == 4)
				 || (st_pa_carr.odd == 1 && st_pa_carr.carr_prev == 2)) {
					strcpy(bit_data,"?00");
					st_pa_carr.carr_prev = -4;
				} else {
					strcpy(bit_data,"?0");
					st_pa_carr.carr_prev = -2;
				}
			} else {
				if ((st_pa_carr.odd == 0 && st_pa_carr.carr_prev == -4)
				 || (st_pa_carr.odd == 1 && st_pa_carr.carr_prev == -2)) {
					strcpy(bit_data,"?11");
					st_pa_carr.carr_prev = 4;
				} else {
					strcpy(bit_data,"?1");
					st_pa_carr.carr_prev = 2;
				}
			}
			st_lamda.us_mid_avg[fsk_spd] = (st_lamda.us_mid_avg[fsk_spd] + lamda) / 2.0;

			rep_p1_sample_num[2]++;
		}
		else if (st_lamda.us_max[fsk_spd] < lamda) {
			// too long
			if (hi) {
				strcpy(bit_data,"?LL");
				st_pa_carr.carr_prev = -5;
			} else {
				strcpy(bit_data,"?HH");
				st_pa_carr.carr_prev = 5;
			}
			rep_p1_sample_num[3]++;
		}
		else if (lamda < st_lamda.us_min[fsk_spd+1]) {
			// too short
			if (hi) {
				strcpy(bit_data,"?L");
				st_pa_carr.carr_prev = -1;
			} else {
				strcpy(bit_data,"?H");
				st_pa_carr.carr_prev = 1;
			}
			rep_p1_sample_num[4]++;
		}
		else {
			// error
			strcpy(bit_data,"??");
			st_pa_carr.carr_prev = 0;

			rep_p1_sample_num[5]++;
		}
	}
	// データ追加
	if (bit_data[0] != '?') {
		// OK data
		c_data->adds(bit_data, ms);

		if (param.GetDebugMode() > 1) {
			// デバッグログ
			sprintf(cbuff, "p1 w:%12d(%s) c:%12d: %2d %3.6f %s\n"
				, w_data->total_r_pos
				, UTILS::get_time_cstr(ms)
				, c_data->total_w_pos
				, st_pa_carr.sample_cnt
				, lamda
				, bit_data);
				if (fiolog) fwrite((void *)cbuff, strlen(cbuff), 1, fiolog);
		}
	} else {
		// NG data
		c_data->adds(&bit_data[1], ms);

		if (param.GetDebugMode() > 1) {
			// デバッグログ
			sprintf(cbuff, "p1 w:%12d(%s) c:%12d: %2d %3.6f %s\n"
				, w_data->total_r_pos
				, UTILS::get_time_cstr(ms)
				, c_data->total_w_pos
				, st_pa_carr.sample_cnt
				, lamda
				, bit_data);
				if (fiolog) fwrite((void *)cbuff, strlen(cbuff), 1, fiolog);
		}
	}

	st_pa_carr.wav_prev = data;
	st_pa_carr.sample_cnt = 0;

	return data;
}

/// @brief 一致するパターンを探す
///
/// @param[in]   c_data 搬送波
/// @param[out]  s_data シリアルデータ
/// @param[in]   baud   600baud:0 1200baud:1 2400baud:2 300baud:3
/// @param[out]  frip   1->0->1->0: 0 0->1->0->1:1
///   
/// @return スタート位置 / -1 なし
int PARSEWAV::find_start_carrier_bit(Buf_t<char> *c_data, Buf_t<char> *s_data, int baud, int *frip)
{
	int pos1[4];
	int pos = -1;
	int idx_ptn = (baud & 3);
	int best_num = -1;
	int best_pos = (int)c_data->size;
	int i;
//	int ptn;
    char bit[4];
	uint32_t ms;

	strcpy(bit, "?");

	// 2400ボーの時はエッジを探す
	if (idx_ptn == IDX_PTN_2400) {
		for(i=0; i<4; i++) {
			pos1[i]=c_data->find(0, carrier_edge_pattern[i]);
		}
		for(i=0; i<4; i++) {
			if (pos1[i] >= 0 && best_pos > pos1[i]) {
				best_pos = pos1[i];
				best_num = i;
			}
		}
		if (best_num >= 0) {
			// あり
			pos = best_pos + (int)strlen(carrier_edge_pattern[best_num]);

			if ((best_num % 2) == 0) {
				(*frip) = 0;
			} else {
				(*frip) = 1;
			}
			if ((best_num / 2) == 0) {
				strcpy(bit, "01");
			} else {
				strcpy(bit, "10");
			}
		}
	}

	// 一致するパターンを探す
	if (best_num < 0) {
		for(i=0; i<4; i++) {
			pos1[i]=c_data->find(0, carrier_pattern[idx_ptn][i]);
		}
		for(i=0; i<4; i++) {
			if (pos1[i] >= 0 && best_pos > pos1[i]) {
				best_pos = pos1[i];
				best_num = i;
			}
		}
		if (best_num >= 0) {
			// あり
			pos = best_pos + (int)strlen(carrier_pattern[idx_ptn][best_num]);

			if ((best_num % 2) == 0) {
				(*frip) = 0;
			} else {
				(*frip) = 1;
			}
			if ((best_num / 2) == 0) {
				strcpy(bit, "0");
			} else {
				strcpy(bit, "1");
			}
		}
	}

	if (pos >= 0) {
		// データ有り
		ms = c_data->get_ms(best_pos);

		if (param.GetDebugMode() > 1) {
			// デバッグログ
			sprintf(cbuff, "p2 fst c:%12d(%s) w_pos:%d pos:%d frip:%d ("
				, c_data->total_r_pos
				, UTILS::get_time_cstr(ms)
				, s_data->total_w_pos
				, best_pos, *frip);
			strncat(cbuff, c_data->curptr(best_pos), pos - best_pos);
			strcat(cbuff, ")\n");
				if (fiolog) fwrite((void *)cbuff, strlen(cbuff), 1, fiolog);
		}

		c_data->r_pos += pos;
		c_data->total_r_pos += pos;

		s_data->adds(bit, ms);
	} else {
		// データなし
		c_data->total_r_pos += c_data->w_pos - c_data->r_pos;
		c_data->r_pos = c_data->w_pos;
	}

	return pos;
}

/// @brief 搬送波(2400/1200Hz)のデータからシリアルデータを１ビット取り出す
///
/// @param[in]  c_data  搬送波
/// @param[out] s_data    シリアルデータ
/// @param[in]  baud    600baud:0 1200baud:1 2400baud:2 300baud:3
/// @param[in,out]  frip    1->0->1->0: 0 0->1->0->1:1

/// @return 次の位置
int PARSEWAV::decode_to_serial_data(Buf_t<char> *c_data, Buf_t<char> *s_data, int baud, int *frip)
{
	int pos = -1;
	int idx_ptn = (baud & 3);
	int len;
//	int ptn;
	char bit_data[4];
	uint32_t ms;
	const char *pptn;

	strcpy(bit_data, "");

	len = (int)strlen(carrier_pattern[idx_ptn][*frip]);
	pptn = carrier_pattern[idx_ptn][*frip];
	if (c_data->cmp(0, pptn, len) == 0) {	// 0
		// 0
		strcpy(bit_data, "0");
		pos = len;
		if (idx_ptn == IDX_PTN_2400) {
			// 2400ボーで0の場合、常にfripする
			*frip = (1 - *frip);
		}
	}
	if (pos < 0) {
		len = (int)strlen(carrier_pattern[idx_ptn][2 + *frip]);
		pptn = carrier_pattern[idx_ptn][2 + *frip];
		if (c_data->cmp(0, pptn, len) == 0)	{ // 1
			// 1
			strcpy(bit_data, "1");
			pos = len;
		}
	}

	ms = c_data->get_ms(0);

	if (pos >= 0) {
		// データ追加
		s_data->adds(bit_data, ms);

		c_data->r_pos += pos;
		c_data->total_r_pos += pos;
	} else {
		// 一致するパターンがない
		strcpy(bit_data, "?");
		s_data->adds(bit_data, ms);

		if (c_data->sublen() > 0) {
			c_data->r_pos += 1;
			c_data->total_r_pos += 1;
		}

		rep_p2_error_num++;

		if (param.GetDebugMode() > 1) {
			// デバッグログ
			sprintf(cbuff,"p2 c2s c:%12d(%s) s:%8d: error ("
				, c_data->total_r_pos
				, UTILS::get_time_cstr(ms)
				, s_data->total_w_pos
				);
			strncat(cbuff, c_data->curptr(0)
				, c_data->sublen() >= len ? len : c_data->sublen());
			strcat((char *)cbuff, ")\n");
				if (fiolog) fwrite((void *)cbuff, strlen(cbuff), 1, fiolog);
		}
	}	

	return pos;
}

/// @brief 搬送波からボーレートを判定する
///
/// @param[in] fsk_spd 1:倍速FSK
/// @param[in] c_data 搬送波
/// @return 解析した位置 / -1:該当なし
int PARSEWAV::parse_carrier_pattern(int fsk_spd, Buf_t<char> *c_data)
{
	int len = 0;
	int pos = -1;
	int pos0 = 0;
	int pos1 = 0;
	int idx_ptn;
	int idx;
	char buf0[16 * 11];
	char buf1[16 * 11];

	for(idx = 0; idx < 4; idx++) {
		idx_ptn = c_baud_min[idx];
		// 0になる位置を集計
		strcpy(buf0, carrier_pattern[idx_ptn][2]);	// 1
		strcpy(buf1, carrier_pattern[idx_ptn][3]);
		strcat(buf0, carrier_pattern[idx_ptn][0]);	// 0
		strcat(buf1, carrier_pattern[idx_ptn][1]);
		int idx_frip = (idx_ptn == 2 ? 3 : 2);	// 2400ボーの時は逆
		for(int i=0; i<8; i++) {
			strcat(buf0, carrier_pattern[idx_ptn][idx_frip]);	// 11111111
			strcat(buf1, carrier_pattern[idx_ptn][5 - idx_frip]);
		}
		pos0=c_data->find(0, buf0);
		pos1=c_data->find(0, buf1);
		if (pos0 >= 0) {
			// 順波
			st_chkwav[fsk_spd].rev_num[st_chkwav[fsk_spd].num][0]++;
		}
		if (pos1 >= 0) {
			// 逆波
			st_chkwav[fsk_spd].rev_num[st_chkwav[fsk_spd].num][1]++;
		}
		if (pos0 >= 0 || pos1 >= 0) {
			len = (int)strlen(buf0);
			st_chkwav[fsk_spd].baud_num[st_chkwav[fsk_spd].num][idx_ptn]++;
			break;
		}
	}

	if (pos0 >= 0 || pos1 >= 0) {
		if (pos0 >= pos1) {
			pos = pos0 + len;
		} else if (pos1 >= pos0) {
			pos = pos1 + len;
		}
	}
	return pos;
}

/// @brief ボーレートを判定してシリアルデータを変換
/// 元のシリアルデータが2400ボーで解析しているのでヘッダ($FF)のデータ長をもとに
/// ボーレートを切り替える
///
/// @param[in] s_data  シリアルデータ
/// @param[out] sn_data 変換後シリアルデータ
/// @return 意味を持たない
int PARSEWAV::conv_baud_rate(Buf_t<char> *s_data, Buf_t<char> *sn_data)
{
	int pos = 0;
	uint32_t ms;
	uint8_t baud;
	int sn_w_pos;
	int s_r_pos;
	int idx;
	int o_idx;
	int cnt;
	int prev_bcnt = phase2n_baud_count;	// for debug
	char buf_s[100];
	int baud_mag = param.GetFskSpeed() + 1;

	// ボーレート自動判定
	if (param.GetAutoBaud() == 1) {
		s_r_pos = s_data->total_r_pos;
		sn_w_pos = sn_data->total_w_pos;

		bool match = false;
		for(idx = 0; idx < 4 && !match; idx++) {
			if (s_data->cmp(0, conv_baud_tbl_ff[idx].ptn, conv_baud_tbl_ff[idx].len) == 0) {
				match = true;
				if ((phase2n_baud_count & 0xff00) == 0) phase2n_baud_count = ((idx+1) << 8);
				pos = conv_baud_tbl_ff[idx].len;
				o_idx=((phase2n_baud_count & 0xff00) >> 8)-1;
				cnt = (phase2n_baud_count & 0xff);
				if (o_idx >= 0 && o_idx != idx) {
					//　別のボーレートの分をクリア
					for(int i=0; i<cnt; i++) {
						ms = s_data->get_ms(0) - (i * 110 / conv_baud_tbl_ff[o_idx].baud);
						baud = c_baud_min[o_idx] + 1;
						sn_data->addns("01111111111", ms, baud, 11);
					}
					phase2n_baud_count = ((idx+1) << 8);
				}
				phase2n_baud_count++;
				cnt = (phase2n_baud_count & 0xff);
				if (cnt > 4) {
					// 連続していたらボーレート切り替え
					phase3_baud = c_baud_min[idx];
					for(int i=0 ; i<cnt; i++) {
						ms = s_data->get_ms(0) + (i * 110 / conv_baud_tbl_ff[idx].baud);
						baud = phase3_baud + 1;
						sn_data->addns("01111111111", ms, baud, 11);
					}
					phase2n_baud_count = 0;
				}
			}
		}
		if (!match) {
			// クリア
			idx=((phase2n_baud_count & 0xff00) >> 8)-1;
			cnt = (phase2n_baud_count & 0xff);
			if (idx >= 0) {
				for(int i=0; i<cnt; i++) {
					ms = s_data->get_ms(0) - (i * 110 / ((int)c_baud_rate[phase3_baud] * baud_mag));
					int n_idx = idx;
					switch(phase3_baud) {
					case 3:
						// 300
						n_idx = idx + 3;
						break;
					default:
						n_idx = idx + (2 - phase3_baud);
						break;
					}
					if (n_idx < 4) {
						baud = phase3_baud + 1;
						sn_data->addns(conv_baud_tbl_ff[n_idx].ptn, ms, baud, conv_baud_tbl_ff[n_idx].len);
					}
				}
			}
			phase2n_baud_count = 0;
		}
		// $FF以外の値
		if (pos == 0) {
			idx = c_baud_ridx[phase3_baud];
			// 300ボーの場合は8バイトごと
			// 600ボーの場合は4バイトごと
			// 1200ボーの場合は2バイトごと
			// 2400ボーの場合は1バイトごと
			cnt = 1 << (3 - idx);
			if (s_data->same_as(0, cnt, s_data->get(0))) {
				baud = phase3_baud + 1;
				sn_data->add(s_data->get(0), s_data->get_ms(0), baud);
				pos = cnt;
			} else {
				pos = 1;
			}
		}
		s_data->r_pos+=pos;
		s_data->total_r_pos+=pos;

		if (param.GetDebugMode() > 0) {
			s_r_pos = s_data->total_r_pos - s_r_pos;
			sn_w_pos = sn_data->total_w_pos - sn_w_pos;
			memset(buf_s, 0, sizeof(buf_s));
			strncpy(buf_s, &s_data->data[s_data->r_pos - s_r_pos], s_r_pos);
			int df = strcmp(buf_s, &sn_data->data[sn_data->w_pos - sn_w_pos]);
			sprintf(cbuff, "p2n cnv b:%d cnt:%04x df:%2d s:%2d:%s -> sn:%2d:%s\n"
				,phase3_baud, prev_bcnt, df, s_r_pos
				, buf_s
				, sn_w_pos
				, &sn_data->data[sn_data->w_pos - sn_w_pos]
				);
			if (fiolog) fwrite((void *)cbuff, strlen(cbuff), 1, fiolog);
		}

	} else {
		// go through
		baud = param.GetBaud() + 1;
		sn_data->add(s_data->get(0), s_data->get_ms(0), baud);
		s_data->r_pos++;
		s_data->total_r_pos++;

	}
	return pos;
}

/// @brief スタートビット（と思われる）部分を探す
///
/// @param[in] s_data シリアルデータ
/// @return 1:スタートビット 0:?
int PARSEWAV::find_start_serial_bit(Buf_t<char> *s_data)
{
	int pos = 0;
	if ((s_data->get(0) & 1) == 0) {
		pos = 1;
	} else {
		if (param.GetDebugMode() > 1) {
			sprintf(cbuff, "p3 s:% 8d %6dms: skip [%c]\n"
				, s_data->total_r_pos
				, s_data->get_ms(0)
				, s_data->get(0));
			if (fiolog) fwrite((void *)cbuff, strlen(cbuff), 1, fiolog);
		}
		s_data->r_pos++;
		s_data->total_r_pos++;
	}

	return pos;
}

/// @brief シリアルデータをバイナリデータに変換する
///
/// @param[in] s_data  シリアルデータ
/// @param[out] b_data バイナリデータ
/// @return 次の位置
int PARSEWAV::decode_to_binary_data(Buf_t<char> *s_data, Buf_t<uint8_t> *b_data)
{
	int pos;
	int bin_data;
	int h_count = 0;	// H bitcount (use for parity check)
	uint32_t ms = 0;
	char *p;


	p = (char *)&(s_data->data[s_data->r_pos]); // 続きから
	bin_data = 0;
	pos = 0;
	ms = s_data->get_ms(0);
	// start bit
	pos++;
	// data bit
	for(; pos < 8 + ((param.GetWordSelect() & 0x04) ? 1 : 0); pos++) {
		bin_data |= (p[pos] & 1) << (pos - 1);
		h_count += (p[pos] & 1);
	}
	// parity bit
	if ((param.GetWordSelect() & 0x07) != 0x04 && (param.GetWordSelect() & 0x07) != 0x05) {
		if (param.GetWordSelect() & 0x01) {
			// odd parity
			if (p[pos] == (h_count & 0x01)) bin_data |= 0xff00;		// error
		} else {
			// even parity
			if (p[pos] == (1 - (h_count & 0x01))) bin_data |= 0xff00;	// error
		}
		pos++;
	}
	// stop bit
	if ((p[pos] & 1) == 0) bin_data |= 0xff00;		// error
	pos++;
	if ((param.GetWordSelect() & 0x07) == 0x04 || (param.GetWordSelect() & 0x06) == 0x00) {
		// 2 stop bit のエラーは無視
//			if ((p[pos] & 1) == 0) bin_data |= 0xff00;		// error
		if ((p[pos] & 1) != 0) pos++;
	}

	// データ追加
	if ((bin_data & 0xff00) == 0) {
		uint8_t attr = 0;
		if (infile_type <= L3C) {
			// baud select
			attr = param.GetAutoBaud() ? s_data->get_attr(0) : param.GetBaud() + 1;
		}
		b_data->add(bin_data & 0xff, ms, attr);
	} else {
		// エラーの場合
		rep_p3_error_num++;
		if (param.GetDebugMode()) {
			// デバッグログ
			sprintf(cbuff,"p3 s:%8d(%s) baud error.\n"
				, s_data->total_r_pos
				, UTILS::get_time_cstr(ms)
				);
			if (fiolog) fwrite((void *)cbuff, strlen(cbuff), 1, fiolog);
		}
	}
	s_data->r_pos += pos;
	s_data->total_r_pos += pos;

	return pos;
}

/// @brief バイナリデータをシリアルデータに変換する
///
/// @param[in] bin_data バイナリデータ1バイト
/// @param[out] s_data  シリアルデータ
/// @return シリアルデータに変換した長さ
int PARSEWAV::encode_to_serial_data(uint8_t bin_data, Buf_t<char> *s_data)
{
	int pos = 0;
	int h_count = 0;	// H bitcount (use for parity check)
	char val;
	char *p;

	p = (char *)&(s_data->data[s_data->w_pos]);	// 続きから

	// start bit
	p[0] = '0';
	// data bit
	for(pos = 0; pos < 7 + ((param.GetWordSelect() & 0x04) ? 1 : 0); pos++) {
		val = (bin_data & (1 << pos)) ? '1' : '0';
		p[pos+1] = val;
		h_count += (val == '1') ? 1 : 0;
	}
	pos++;
	// parity bit
	if ((param.GetWordSelect() & 0x07) != 0x04 && (param.GetWordSelect() & 0x07) != 0x05) {
		if (param.GetWordSelect() & 0x01) {
			// odd parity
			val = (h_count & 0x01) ? '1' : '0';
		} else {
			// even parity
			val = (h_count & 0x01) ? '0' : '1';
		}
		p[pos] = val;
		pos++;
	}
	// stop bit
	val = '1';
	p[pos] = val;
	pos++;
	if ((param.GetWordSelect() & 0x07) == 0x04 || (param.GetWordSelect() & 0x06) == 0x00) {
		// 2 stop bit
		p[pos] = val;
		pos++;
	}

	return pos;
}

/// @brief シリアルデータを搬送波ビットデータに変換する
///
/// @param[in] s_data シリアルデータ
/// @param[out] c_data 搬送波データ
/// @return 搬送波データに変換した長さ
int PARSEWAV::encode_to_carrier_data(Buf_t<char> *s_data, Buf_t<char> *c_data)
{
	int idx_ptn = (param.GetBaud() & 3);
	int len = 0;
	char data;
	uint32_t ms;

	s_data->get(data, ms);
	if (data & 0x01) {
		// 1
		len = c_data->adds(carrier_pattern[idx_ptn][2 + baud24_frip], ms);
	} else {
		// 0
		len = c_data->adds(carrier_pattern[idx_ptn][baud24_frip], ms);
		if (idx_ptn == IDX_PTN_2400) {
			// 2400ボーの時 fripする
			baud24_frip = (1 - baud24_frip);
		}
	}

	return len;
}

/// @brief 搬送波ビットデータをwaveサンプルデータに変換する
/// @brief (48000Hz 16bit)
///
/// @param[in] c_data 搬送波データ
/// @param[out] w_data waveサンプルデータ
/// @return waveサンプルデータに変換した長さ
int PARSEWAV::encode_to_wave_data(Buf_t<char> *c_data, uint8_t *w_data)
{
	int pos = 0;
	int8_t wav_samples[10];
	int last = param.GetFskSpeed() ? 5 : 10;

	memset(wav_samples, 0x70, 10);

	if (c_data->r_pos <= 0
	|| (c_data->data[c_data->r_pos-1] & 1) != (c_data->data[c_data->r_pos] & 1)) {
		wav_samples[0] = 0x6c;
	}
	if (c_data->r_pos >= c_data->w_pos - 1
	|| (c_data->data[c_data->r_pos] & 1) != (c_data->data[c_data->r_pos+1] & 1)) {
		wav_samples[last-1] = 0x6c;
	}

	// convert to wave sample data
	for(int i=0; i<last; i++) {
		w_data[pos++]=0x80 + ((c_data->data[c_data->r_pos] & 1) ? wav_samples[i] : - wav_samples[i]);
	}
	return pos;
}


/// @brief waveサンプルデータから搬送波(1200/2400Hz)を解析し
/// @brief ビットデータに変換する。(FSK)
///
/// @param[in]  fsk_spd 1:倍速FSK
/// @param[in]  w_data waveデータ
/// @param[out] wc_data wave補正データ
/// @param[out] c_data 搬送波データ
/// @param[out] s_data シリアルデータ
/// @param[out] sn_data 変換後シリアルデータ
/// @param[out] b_data バイナリデータ
/// @param[in]  start_phase 開始フェーズ
/// @return フェーズ番号
int PARSEWAV::decode_phase1(int fsk_spd, Buf_t<uint8_t> *w_data, Buf_t<uint8_t> *wc_data, Buf_t<char> *c_data, Buf_t<char> *s_data, Buf_t<char> *sn_data, Buf_t<uint8_t> *b_data, int start_phase)
{
	return decode_phase1(fsk_spd, w_data, wc_data, c_data, s_data, sn_data, b_data, start_phase, param.GetCorrectType(), param.GetReverseWave());
}

/// @brief waveサンプルデータから搬送波(1200/2400Hz)を解析し
/// @brief ビットデータに変換する。(FSK)
///
/// @param[in]  fsk_spd 1:倍速FSK
/// @param[in]  w_data waveデータ
/// @param[out] wc_data wave補正データ
/// @param[out] c_data 搬送波データ
/// @param[out] s_data シリアルデータ
/// @param[out] sn_data 変換後シリアルデータ
/// @param[out] b_data バイナリデータ
/// @param[in]  start_phase 開始フェーズ
/// @param[in]  correct_type 補正方法
/// @param[in]  reverse 波形反転
/// @return フェーズ番号
int PARSEWAV::decode_phase1(int fsk_spd, Buf_t<uint8_t> *w_data, Buf_t<uint8_t> *wc_data, Buf_t<char> *c_data, Buf_t<char> *s_data, Buf_t<char> *sn_data, Buf_t<uint8_t> *b_data, int start_phase, int correct_type, int reverse)
{
//	int phase1h_frip = 0;
//	bool last = false;
//	bool break_data = false;
	long pos = 0;
	Buf_t<uint8_t> *wn_data;

	init_parameters();

	if (infile_type == WAV && (param.GetCorrectWave() == 1 || chk_wav_mode == true)) {
		wn_data = wc_data;
		dft.Init(st_lamda.samples[fsk_spd], correct_type);
	} else {
		wn_data = w_data;
	}


	while(phase1 >= 0) {
		if (chk_wav_mode == true) {
			pos = (int)(sample_pos * 100 / st_chkwav[fsk_spd].analyze_num / 4) + (int)(st_chkwav[fsk_spd].num * 100 / 4) + (fsk_spd * 50);
		} else {
			pos = (int)(sample_pos * 100 / sample_num);
		}
		if (progbox->setProgress((int)pos)) {
			phase1 = -1;
			break;
		}
		switch(phase1) {
			case 0:
				// WAVファイル読み込み
				get_wav_sample(w_data, reverse);
				// チェックモードの場合は約30秒まで解析
				if (chk_wav_mode == true && sample_pos >= st_chkwav[fsk_spd].analyze_num) {
					w_data->last = true;
				}
				phase1 = 1;
				break;
			case 1:
				// WAVサンプルデータを補正する
				if (infile_type == WAV && (param.GetCorrectWave() == 1 || chk_wav_mode == true)) {
					dft.Calcrate(w_data, wc_data);
				}
				if (outfile_type == WAV && outfile_type >= infile_type) {
					// WAV ファイル出力
					conv.OutConvSampleData((void *)wn_data->startptr(), inwav.GetSampleRate(), sizeof(uint8_t), wn_data->length(), outwav.GetSampleRate(), outwav.GetBlockSize(), fio_out);
				}
				// データ最後まで読んだ場合
				if (param.GetCorrectWave() == 1 || chk_wav_mode == true) {
					if (w_data->last && w_data->isEOB(st_lamda.samples[0])) {
						wc_data->last = true;
					}
				}

				phase1 = 2;
				break;
			case 2:
				// L3Cへ変換
				decode_to_carrier_wave_bn(fsk_spd, wn_data, c_data);
				if (c_data->isFull(3)) {
					// cバッファがいっぱいになったら吐き出す
					phase1 = 21;
				} else if (wn_data->last != true && wn_data->isEOB(st_lamda.samples[0])) {
					// cバッファがいっぱいになるまで繰り返す
					phase1 = 31;
				} else if (wn_data->last && wn_data->isEOB(0)) {
					// 最後のデータ
					c_data->last = true;
					phase1 = 21;
				}
				break;
			case 11:
				// L3Cファイル読み込み
				get_l3c_sample(c_data);
				phase1 = 21;
				break;
			case 21:
				if (outfile_type == L3C && outfile_type >= infile_type) {
					// L3C ファイル出力
					write_pos = UTILS::write_data(fio_out, c_data->startptr(), c_data->length(), 88, write_pos, 2);
				}
				// 更に変換(L3C -> L3B -> L3)
				if (outfile_type >= WAV) {
					if (chk_wav_mode == true) {
						phase2 = 11;
						decode_phase2(fsk_spd, c_data, s_data, sn_data, b_data, 11);
					} else {
						decode_phase2(fsk_spd, c_data, s_data, sn_data, b_data, 1);
					}
				}
				phase1 = 31;
				break;
			case 31:
				// 残りをバッファの最初にコピー→次のターンで解析をするため
				w_data->shift();
				wc_data->shift();

				if (c_data->last && c_data->isEOB(0)) {
					phase1 = -1;
					break;
				}

				if (chk_wav_mode == true && w_data->last == true) {
					phase1 = 41;
				} else {
					phase1 = start_phase;
				}
				break;
			case 41:
				phase1 = -1;
				break;
		}
	}

	return phase1;
}

/// @brief 搬送波(1200/2400Hz)データからシリアルデータに変換
/// ここでボーレートによってシリアルデータが変わってくる
///
/// @param[in]  fsk_spd 1:倍速FSK
/// @param[in] c_data  搬送波データ
/// @param[out] s_data シリアルデータ
/// @param[out] sn_data 変換後シリアルデータ
/// @param[out] b_data バイナリデータ
/// @param[in] start_phase 開始フェーズ
/// @return フェーズ番号
int PARSEWAV::decode_phase2(int fsk_spd, Buf_t<char> *c_data, Buf_t<char> *s_data, Buf_t<char> *sn_data, Buf_t<uint8_t> *b_data, int start_phase)
{
	int step = 0;
	bool break_data = false;
	int baud = (param.GetAutoBaud() ? 2 : param.GetBaud());	///< 自動判定のときは2400ボーにする
	int pos;

	while(phase2 >= 0) {
		if (chk_wav_mode == true) {
			pos = (int)(sample_pos * 100 / st_chkwav[fsk_spd].analyze_num / 4) + (int)(st_chkwav[fsk_spd].num * 100 / 4) + (fsk_spd * 50);
		} else {
			pos = (int)(sample_pos * 100 / sample_num);
		}
		if (progbox->setProgress(pos)) {
			phase1 = -1;
			break;
		}
		// バッファの末尾
		if (c_data->last != true && c_data->isEOB(32)) {
			break_data = true;
		}
		if (break_data == true) {
			break;
		}
		switch (phase2) {
			case 0:
				step = find_start_carrier_bit(c_data, s_data, baud, &phase2_frip);
				if (c_data->last && c_data->isEOB(0)) {
					s_data->last = true;
					phase2 = 31;
					break;
				}
				if (step >= 0) {
					if (s_data->isFull(0)) {
						phase2 = 31;
						break;
					}
					phase2 = 1;
				} else {
					break_data = true;
				}
				break;
			case 1:
				step = decode_to_serial_data(c_data, s_data, baud, &phase2_frip);
				if (s_data->isFull(0)) {
					phase2 = 31;
					break;
				}
				if (c_data->isEOB(0)) {
					s_data->last = true;
					phase2 = 31;
					break;
				}
				if (step < 0) {
					phase2 = 0;
				}
				break;
			case 11:
				// 搬送波からボーレートを解析
				step = parse_carrier_pattern(fsk_spd, c_data);
				if (step >= 0) {
					c_data->r_pos += step;
					c_data->total_r_pos += step;
				} else {
					// no data found or parsed all c_data
					c_data->total_r_pos += c_data->w_pos;
					c_data->clear();
					phase2 = -1;
				}
				break;
			case 21:
				get_l3b_sample(s_data);
				phase2 = 31;
				break;
			case 22:
				get_t9x_sample(s_data);
				phase2 = 31;
				break;
			case 31:
//				if (outfile_type == L3B && outfile_type >= infile_type) {
//					write_pos = write_data(fio_out, s_data->startptr(), s_data->length(), 110, write_pos, 1, 1);
//				}
				if (outfile_type >= WAV) decode_phase2n(s_data, sn_data, b_data, 0);
				if (s_data->last && s_data->isEOB(0)) {
					phase2 = -1;
					break;
				}
				phase2 = start_phase;
				break;

		}
	}
	if (c_data->last != true && break_data == true) {
		// 残りをバッファの最初にコピー→次のターンで解析をするため
		c_data->shift();
	}
	return phase2;
}

/// @brief L3B -> ボーレート変換 ->  L3B
///
/// @param[in] s_data シリアルデータ
/// @param[out] sn_data 変換後シリアルデータ
/// @param[out] b_data バイナリデータ
/// @param[in] start_phase 開始フェーズ
/// @return フェーズ番号
int PARSEWAV::decode_phase2n(Buf_t<char> *s_data, Buf_t<char> *sn_data, Buf_t<uint8_t> *b_data, int start_phase)
{
	int step = 0;
//	int baud = 0;
	bool break_data = false;

	while(phase2n >= 0) {
		if (progbox->setProgress((int)(sample_pos * 100 / sample_num))) {
			phase2n = -1;
			break;
		}
		// バッファの末尾
		if (s_data->last != true && s_data->isEOB(32)) {
			break_data = true;
		}
		if (break_data == true) {
			break;
		}
		switch (phase2n) {
			case 0:
				step = conv_baud_rate(s_data, sn_data);
				if (sn_data->isFull(131)) {
					phase2n = 31;
					break;
				}
				if (s_data->isEOB(0)) {
					sn_data->last = true;
					phase2n = 31;
					break;
				}
				break;
			case 21:
				get_l3b_sample(sn_data);
				phase2n = 31;
				break;
			case 22:
				get_t9x_sample(sn_data);
				phase2n = 31;
				break;
			case 31:
				if (outfile_type == L3B && outfile_type >= infile_type) {
					// L3Bファイル出力
					write_pos = UTILS::write_data(fio_out, sn_data->startptr(), sn_data->length(), 110, write_pos, 1);
				}
				else if (outfile_type == T9X && outfile_type >= infile_type) {
					// T9Xファイル出力
					write_pos = UTILS::out_t9x_data(fio_out, sn_data->startptr(), sn_data->length(), write_pos, sn_data->last);
				}
				if (outfile_type >= WAV) decode_phase3(sn_data, b_data, 0);
				if (s_data->last && s_data->isEOB(0)) {
					phase2n = -1;
					break;
				}
				phase2n = start_phase;
				break;

		}
	}
	if (s_data->last != true && break_data == true) {
		// 残りをバッファの最初にコピー→次のターンで解析をするため
		s_data->shift();
	}
	return phase2n;
}

/// @brief L3B -> L3 シリアルからバイナリに変換
///
/// @param[in] s_data シリアルデータ
/// @param[out] b_data バイナりデータ
/// @param[in] start_phase 開始フェーズ
/// @return フェーズ番号
int PARSEWAV::decode_phase3(Buf_t<char> *s_data, Buf_t<uint8_t> *b_data, int start_phase)
{
	int step = 0;
	bool break_data = false;

	while(phase3 >= 0) {
		if (progbox->setProgress((int)(sample_pos * 100 / sample_num))) {
			phase3 = -1;
			break;
		}
		// バッファの末尾
		if (s_data->last != true && s_data->isEOB(32)) {
			break_data = true;
		}
		if (break_data == true) {
			break;
		}
		switch (phase3) {
			case 0:
				// find start bit
				step = find_start_serial_bit(s_data);
				if (step > 0) {
					phase3 = 1;
				}
				if (s_data->isEOB(0)) {
					b_data->last = true;
					phase3 = 41;
					break;
				}
				break;
			case 1:
				step = decode_to_binary_data(s_data, b_data);
				if (b_data->isFull(0) || s_data->isEOB(0)) {
					phase3 = 41;
					if (s_data->isEOB(0)) {
						b_data->last = true;
					}
					break;
				}
				phase3 = 0;
				break;
			case 31:
				get_l3_sample(b_data);
				phase3 = 41;
				break;
			case 41:
				if (outfile_type == L3 && outfile_type >= infile_type) {
					// L3ファイル出力
					if (b_data->length() > 0) {
						fwrite((void *)b_data->startptr(), sizeof(uint8_t), b_data->length(), fio_out);
					}
				}
				if (outfile_type >= WAV) decode_phase4(b_data);
				if (b_data->last && b_data->isEOB(0)) {
					phase3 = -1;
					break;
				}

				phase3 = start_phase;
				break;
		}
	}
	if (s_data->last != true && break_data == true) {
		// 残りをバッファの最初にコピー→次のターンで解析をするため
		s_data->shift();
	}
	return phase3;
}

/// @brief L3→実データ変換
///
/// @param[in] b_data バイナリデータ
/// @return フェーズ番号
///
int PARSEWAV::decode_phase4(Buf_t<uint8_t> *b_data)
{
//	uint8_t *p = b_data->data;
	int header_type = -1;
	int data_len = 0;
	uint8_t data;
	uint8_t data_body[257];
	uint8_t *pdata;
	uint32_t ms;
	int chk_sum_calc = 0;
	int chk_sum_data = 0;
	int start_pos = 0;
	uint32_t start_ms = 0;
	pos_t err_pos;
	bool break_data = false;

	// ヘッダ解析
//	phase4 = 0;
	while(phase4 >= 0) {
		// バッファの末尾か？
		if (b_data->last != true && b_data->isEOB(16)) {
			break_data = true;
		} else if (b_data->last == true && b_data->isEOB(0)) {
			phase4 = -1;
			break_data = true;
		}
		if (break_data == true) {
			break;
		}

		switch(phase4) {
			case 0:
				// 0xff 0x01 0x3c を探す
				if(b_data->cmp(0, cmt_ident, 3) == 0) {
					// found
					if (rep4_itm == NULL) rep4_itm = new REPORT4();
//					rep_p4_header_num++;
					// type?
					header_type = b_data->get(3);
					switch(header_type) {
						case 0:
							// file name section
							phase4 = 10;
							break;
						case 1:
							// body data section
							phase4 = 11;
							break;
						case 0xff:
							// footer data section
							phase4 = 21;
							break;
						default:
							// ?? error?
							break;
					}
					b_data->r_pos += 4;
					b_data->total_r_pos += 4;
				} else {
					b_data->r_pos++;
					b_data->total_r_pos++;

				}
				break;
			case 10:
				// file name section
				rep4_itm->flags |= 1;
				rep4_itm->baud = b_data->get_attr(0);
				data_len = b_data->get(0);
				if (b_data->last != true && b_data->isEOB(data_len + 6)) {
					// バッファをフラッシュ
					break_data = true;
					break;
				}
				b_data->get(data, ms);

				start_pos = b_data->total_r_pos;
				start_ms = b_data->get_ms(0);
				chk_sum_calc = 0;

				chk_sum_calc += data_len;
				for(int i=0; i<data_len && i<20; i++) {
					b_data->get(data, ms);
					save_data_name[i] = data;
					chk_sum_calc += data;
				}
				save_data_name[20]='\0';

				memcpy(rep4_itm->save_data_name, save_data_name, 21);

				b_data->get(data, ms);
				chk_sum_data = (int)data;

				chk_sum_calc = (chk_sum_calc & 0xff);

				if (chk_sum_calc != chk_sum_data) {
					// レポート用
					err_pos.start_pos = start_pos;
					err_pos.start_ms  = start_ms;
					err_pos.end_pos = b_data->total_r_pos;
					err_pos.end_ms  = b_data->get_ms(1);
					(rep4_itm->chksum_err_pos).push_back(err_pos);
					rep4_itm->chksum_err_num++;
				}

				// open file
				if (outfile_type == REAL && outfile_type >= infile_type) {
					if (param.GetFileSplit()) {
						outsfile.Printf(_T("_%03d"), (int)rep4.size()+1);
						outsfile = outsfileb + outsfile;
						outsfile += outsext;
						if (fio_outs != NULL) fclose(fio_outs);
						fio_outs = UTILS::pw_fopen(outsfile, _T("wb"));
					}
					prev_data_len = 0;
					memset(prev_data_body, 0, sizeof(prev_data_body));
				}

				phase4 = 0;
				break;
			case 11:
				// body data section
				rep4_itm->flags |= 2;
				data_len = b_data->get(0);
				if (b_data->last != true && b_data->isEOB(data_len + 6)) {
					// バッファをフラッシュ
					break_data = true;
					break;
				}
				b_data->get(data, ms);

				start_pos = b_data->total_r_pos;
				start_ms = b_data->get_ms(0);
				chk_sum_calc = 1;

				rep4_itm->data_count++;

				chk_sum_calc += data_len;
				for(int i=0; i<data_len && i<255; i++) {
					b_data->get(data, ms);
					data_body[i] = data;
					chk_sum_calc += data;
				}
				data_body[255]='\0';

				b_data->get(data, ms);
				chk_sum_data = (int)data;

				chk_sum_calc = (chk_sum_calc & 0xff);

				if (chk_sum_calc != chk_sum_data) {
					// レポート用
					err_pos.start_pos = start_pos;
					err_pos.start_ms  = start_ms;
					err_pos.end_pos = b_data->total_r_pos;
					err_pos.end_ms  = b_data->get_ms(1);
					(rep4_itm->chksum_err_pos).push_back(err_pos);
					rep4_itm->chksum_err_num++;
				}

				// write to file
				if (outfile_type == REAL && outfile_type >= infile_type) {
					// マシン語のヘッダを取り除く場合
					if (param.GetDeleteMHead() && rep4_itm->save_data_name[8] == 2 && rep4_itm->data_count == 1) {
						pdata = &data_body[5];
						data_len -= 5;
					} else {
						pdata = data_body;
					}

					if (prev_data_len > 0) {
						if (!param.GetFileSplit()) {
							fwrite(prev_data_body, sizeof(uint8_t), prev_data_len, fio_out);
						} else {
							if (fio_outs != NULL) {
								fwrite(prev_data_body, sizeof(uint8_t), prev_data_len, fio_outs);
							}
						}
					}
					prev_data_len = data_len;
					memcpy(prev_data_body, pdata, data_len);
				}

				phase4 = 0;

				break;
			case 21:
				// footer data section
				rep4_itm->flags |= 4;
				data_len = b_data->get(0);
				if (b_data->last != true && b_data->isEOB(data_len + 6)) {
					// バッファをフラッシュ
					break_data = true;
					break;
				}
				b_data->get(data, ms);

				start_pos = b_data->total_r_pos;
				start_ms = b_data->get_ms(0);
				chk_sum_calc = 255;

				chk_sum_calc += data_len;
				for(int i=0; i<data_len && i<255; i++) {
					b_data->get(data, ms);
					chk_sum_calc += data;
				}
				b_data->get(data, ms);
				chk_sum_data = (int)data;

				chk_sum_calc = (chk_sum_calc & 0xff);

				if (chk_sum_calc != chk_sum_data) {
					// レポート用
					err_pos.start_pos = start_pos;
					err_pos.start_ms  = start_ms;
					err_pos.end_pos = b_data->total_r_pos;
					err_pos.end_ms  = b_data->get_ms(1);
					(rep4_itm->chksum_err_pos).push_back(err_pos);
					rep4_itm->chksum_err_num++;
				}

				// write file
				if (outfile_type == REAL && outfile_type >= infile_type) {
					// マシン語のヘッダを取り除く場合
					if (param.GetDeleteMHead() && rep4_itm->save_data_name[8] == 2) {
						prev_data_len -= 5;
					}
					if (prev_data_len > 0) {
						if (!param.GetFileSplit()) {
							fwrite(prev_data_body, sizeof(uint8_t), prev_data_len, fio_out);
						} else {
							if (fio_outs != NULL) {
								fwrite(prev_data_body, sizeof(uint8_t), prev_data_len, fio_outs);
							}
						}
					}
					if (param.GetFileSplit()) {
						// close file
						if (fio_outs != NULL) {
							fclose(fio_outs);
							fio_outs = NULL;
						}
						// 一旦バッファをフラッシュする。
						break_data = true;
					}
				}

				// add report data
				rep4.push_back(rep4_itm);
				rep4_itm = NULL;

				phase4 = 0;	// end -> parse next data

				break;
		}
	}
	if (b_data->last != true && break_data == true) {
		// 残りをバッファの最初にコピー→次のターンで解析をするため
		b_data->shift();
	}
	if (phase4 < 0 && rep4_itm != NULL) {
		// add report data
		rep4.push_back(rep4_itm);
		rep4_itm = NULL;
	}

	return phase4;
}

/// @brief 実ファイル -> L3
///
/// @param[out] b_data バイナリデータ
/// @param[out] s_data シリアルデータ
/// @param[out] c_data 搬送波データ
/// @return フェーズ番号
int PARSEWAV::encode_phase4(Buf_t<uint8_t> *b_data, Buf_t<char> *s_data, Buf_t<char> *c_data)
{
	uint8_t data;
	int hlen = c_gap_length[0];
	int dlen = 0;
	int chk_sum = 0;

	if (param.GetChangeGapSize() != 0 || outfile_type <= L3C) {
		hlen = param.GetBaud();
		if (hlen >= 3) hlen = 0;
		hlen += (param.GetFskSpeed() & 1);
		hlen = c_gap_length[hlen];
	}

	while (phase4 >= 0) {
		switch(phase4) {
			case 0:
				chk_sum = 0;

				b_data->set(0xff, 0, hlen);
				b_data->addns(&cmt_ident[1], 0, 2);
				// header_type
				b_data->add(0, 0);
				chk_sum += 0;
				// 20バイト
				dlen = 20;
				b_data->add((uint8_t)dlen, 0);
				chk_sum += dlen;
				for(int i=0; i<dlen; i++) {
					data = save_data_name[i];
					b_data->add(data, 0);
					chk_sum += data;
				}
				// chk sum
				b_data->add((chk_sum & 0xff), 0);
				// 0 x4
				b_data->set(0, 0, 4);

				phase4 = 1;
				break;

			case 1:
				// body data
				chk_sum = 0;

				b_data->set(0xff, 0, hlen);
				b_data->addns(&cmt_ident[1], 0, 2);
				// header_type
				b_data->add(1, 0);
				chk_sum += 1;
				// 255バイト最大
				if (sample_num - sample_pos >= 255) {
					dlen = 255;
				} else {
					dlen = (int)(sample_num - sample_pos);
				}
				b_data->add((uint8_t)dlen, 0);
				chk_sum += dlen;
				for(int i=0; i<dlen; i++) {
					data = get_rf_sample();
					b_data->add(data, 0);
					chk_sum += data;
				}
				// chk sum
				b_data->add((chk_sum & 0xff), 0);
				// 0 x 4
				b_data->set(0, 0, 4);

				if (save_data_name[9] == 0) {
					// gap reset when it's binary save
					hlen = 10;
				}

				if (sample_pos >= sample_num) {
					phase4 = 2;
				}
				break;
			case 2:
				// footer data
				chk_sum = 0;

				b_data->set(0xff, 0, hlen);
				b_data->addns(&cmt_ident[1], 0, 2);
				// header_type
				b_data->add(0xff, 0);
				chk_sum += 0xff;
				//
				dlen = 0;
				b_data->add((uint8_t)dlen, 0);
				chk_sum += dlen;
				// chk sum
				b_data->add((chk_sum & 0xff), 0);
				// 0 x 4
				b_data->set(0, 0, 4);

				phase4 = -1;
				break;
		}
		if (b_data->isFull(300) || phase4 < 0) {
			if (outfile_type == L3) {
				fwrite(b_data->data, sizeof(uint8_t), (b_data->w_pos), fio_out);
			}
			phase3 = 40;
			if (outfile_type < L3) encode_phase3(b_data, s_data, c_data, 40);
			b_data->clear();
		}
	}

	return phase4;
}

/// @brief L3 -> L3B バイナリからシリアルに変換
///
/// @param[in] b_data バイナリデータ
/// @param[out] s_data シリアルデータ
/// @param[out] c_data 搬送波データ
/// @param[in] start_phase 開始フェーズ
/// @return フェーズ番号
int PARSEWAV::encode_phase3(Buf_t<uint8_t> *b_data, Buf_t<char> *s_data, Buf_t<char> *c_data, int start_phase)
{
//	int pos = 0;
	int step = 0;
	bool last = false;
	uint8_t bin_data = 0;

//	pos = strlen(serial_data);

	while(phase3 >= 0) {
		if (progbox->setProgress((int)(sample_pos * 100 / sample_num))) {
			phase3 = -1;
			break;
		}
		switch (phase3) {
			case 40:
				// バイナリデータ読み込み
				bin_data = b_data->data[b_data->r_pos++];
				if (b_data->isEOB(0)) {
					b_data->last = true;
					last = true;
				}
				phase3 = 31;
				break;
			case 30:
				// l3データ読み込み or PLAINデータ読み込み
				bin_data = get_l3_sample();
				if (sample_pos >= sample_num) {
					b_data->last = true;
					last = true;
				}
				phase3 = 31;
				break;
			case 31:
				// シリアルデータに変換
				step = encode_to_serial_data(bin_data, s_data);
				phase3 = start_phase;
				s_data->w_pos += step;
				if (s_data->isFull(10)) {
					phase3 = 11;
				}
				if (b_data->last) {
					s_data->last = true;
					phase3 = 11;
				}
				break;
			case 20:
				// l3bデータ読み込み
				get_l3b_sample(s_data);
				phase3 = 11;
				if (s_data->last) {
					last = true;
				}
				break;
			case 21:
				// t9xデータ読み込み
				get_t9x_sample(s_data);
				phase3 = 11;
				if (s_data->last) {
					last = true;
				}
				break;
			case 11:
				// 次のフェーズへ
				if (outfile_type == L3B) {
					write_pos = UTILS::write_data(fio_out, s_data->data, s_data->w_pos, 110, write_pos, 1);
					s_data->clear();
				}
				else if (outfile_type == T9X) {
					write_pos = UTILS::out_t9x_data(fio_out, s_data->data, s_data->w_pos, write_pos, s_data->last);
					s_data->clear();
				}
				if (outfile_type < L3B) encode_phase2(s_data, c_data, 21, last);
//				pos = strlen(serial_data);
				if (last) {
					phase3 = -1;
					break;
				}
				phase3 = start_phase;
				break;
		}
	}
	return phase3;

}

/// @brief l3b -> l3c シリアルデータを搬送波ビットデータに変換する
///
/// @param[in] s_data シリアルデータ
/// @param[out] c_data 搬送波データ
/// @param[in] start_phase 開始フェーズ
/// @param[in] end_data 最終データか？
/// @return フェーズ番号
int PARSEWAV::encode_phase2(Buf_t<char> *s_data, Buf_t<char> *c_data, int start_phase, bool end_data)
{
//	int pos = 0;
//	int step = 0;
//	int pos_s = 0;
//	int len = 0;
	bool last = false;
	bool break_data = false;

//	len = strlen(serial_data);
//	pos = strlen(carrier_data);
//	pos = c_data->w_pos;

	while(phase2 >= 0) {
		if (progbox->setProgress((int)(sample_pos * 100 / sample_num))) {
			phase2 = -1;
			break;
		}

		// バッファの末尾
		if (end_data != true && s_data->isEOB(32)) {
			break_data = true;
		}
		if (break_data == true) {
			break;
		}

		switch (phase2) {
			case 21:
				// 搬送波データに変換
				encode_to_carrier_data(s_data, c_data);
				phase2 = start_phase;
				if (c_data->isFull(7)) {
					phase2 = 11;
				}
				if (s_data->isEOB(0)) {
					phase2 = 11;
					last = true;
				}
				break;
			case 10:
				get_l3c_sample(c_data);
				phase2 = 11;
				if (c_data->last) {
					last = true;
				}
				break;
			case 11:
				if (outfile_type == L3C) {
					write_pos = UTILS::write_data(fio_out, c_data->data, c_data->w_pos, 88, write_pos, 2);
					c_data->clear();
				}
				if (outfile_type < L3C) encode_phase1(c_data, last);
				if (last) {
					phase2 = -1;
					break;
				}
				phase2 = start_phase;
				break;
		}
	}
	if (end_data != true && break_data == true) {
		// 残りをバッファの最初にコピー→次のターンで解析をするため
		s_data->shift();
	}
	return phase2;

}

/// @brief L3C -> WAV 搬送波からwaveサンプルデータに変換
///
/// @param[in] c_data 搬送波データ
/// @param[in] end_data 最終データか？
/// @return フェーズ番号
int PARSEWAV::encode_phase1(Buf_t<char> *c_data, bool end_data)
{
	int pos = 0;
	int step = 0;
	bool last = false;
	bool break_data = false;
	uint8_t wav_data[20];

	while(phase1 >= 0) {
		if (progbox->setProgress((int)(sample_pos * 100 / sample_num))) {
			phase1 = -1;
			break;
		}

		// バッファの末尾
		if (end_data != true && c_data->isEOB(32)) {
			break_data = true;
		}
		if (break_data == true) {
			break;
		}

		step = encode_to_wave_data(c_data, wav_data);
		pos += step;
		c_data->r_pos++;
		c_data->total_r_pos++;

		if (c_data->isEOB(0)) {
			last = true;
		}

		if (outfile_type == WAV) {
			conv.OutConvSampleData(wav_data,48000,1,step,outwav.GetSampleRate(),outwav.GetBlockSize(),fio_out);
			pos = 0;
		}

		if (last) {
			phase1 = -1;
			break;
		}

	}

	if (end_data != true && break_data == true) {
		// 残りをバッファの最初にコピー→次のターンで解析をするため
		c_data->shift();
	}
	return phase1;

}

/// @brief レポートログ出力
///
///
///
///
void PARSEWAV::reporting()
{
//	size_t len;
	int spd = param.GetFskSpeed();

	buff = _T("----- Result Report -----");
	write_log(buff, 1);

	// input
	buff = _T(" [ input ]");
	write_log(buff, 1);
	buff = _T(" ") + infile;
	write_log(buff, 1);
	if (infile_type == WAV) {
		buff.Printf(_T(" %dHz %dbit %dch"),inwav.GetSampleRate(), inwav.GetSampleBits(), inwav.GetChannels());
		write_log(buff, 1);
	}

	write_log(_T(""), 1);

	// output
	if (outfile_type != NO_FILE) {
		buff = _T(" [ output ]");
		write_log(buff, 1);
		buff = _T(" ") + outfile;
		write_log(buff, 1);

		write_log(_T(""), 1);
	}
	// phase1 report
	if (infile_type == WAV && outfile_type >= WAV) {
		buff = _T(" [ wav -> l3c ]");
		write_log(buff, 1);
		buff.Printf(_T("  Wave Type: %s"), (spd ? _T("Double Speed FSK") : _T("Standard FSK")));
		write_log(buff, 1);
		buff.Printf(_T("  Detection: %s"),(param.GetHalfWave() ? _T("Half wave") : _T("Full wave")));
		write_log(buff, 1);
		buff.Printf(_T("  Range: Long(0) : %2d%% (%3.3fus(%6.1fHz) - %3.3fus(%6.1fHz) - %3.3fus(%6.1fHz))")
			,param.GetRange(0)
			,st_lamda.us_max[spd],	(1000000.0 / st_lamda.us_max[spd])
			,st_lamda.us[spd],		(1000000.0 / st_lamda.us[spd])
			,st_lamda.us_min[spd],  (1000000.0 / st_lamda.us_min[spd])
		);
		write_log(buff, 1);
		buff.Printf(_T("  Range: Short(1): %2d%% (%3.3fus(%6.1fHz) - %3.3fus(%6.1fHz) - %3.3fus(%6.1fHz))")
			,param.GetRange(1)
			,st_lamda.us_max[spd+1],(1000000.0 / st_lamda.us_max[spd+1])
			,st_lamda.us[spd+1],	(1000000.0 / st_lamda.us[spd+1])
			,st_lamda.us_min[spd+1],(1000000.0 / st_lamda.us_min[spd+1])
		);
		write_log(buff, 1);
		buff.Printf(_T("  Wave Reverse: %s"),(param.GetReverseWave() ? _T("on") : _T("off")));
		write_log(buff, 1);
		buff.Printf(_T("  Correct: %s"),(param.GetCorrectWave() ? _T("on") : _T("off")));
		write_log(buff, 1);
		if (param.GetCorrectWave() == 1) {
			buff.Printf(_T("  Correct Type: %s"),(param.GetCorrectType() ? _T("sin wave") : _T("cos wave")));
			write_log(buff, 1);
		}
		buff = _T("  Wave Sum:");
		write_log(buff, 1);

		buff.Printf(_T("    Long(0) : %8d Cent:%6.1fHz Avg:%6.1fHz")
			, rep_p1_sample_num[0]
			, 1000000.0 / st_lamda.us[spd]
			, 1000000.0 / st_lamda.us_avg[spd]);
		write_log(buff, 1);
		buff.Printf(_T("    Short(1): %8d Cent:%6.1fHz Avg:%6.1fHz")
			, rep_p1_sample_num[1]
			, 1000000.0 / st_lamda.us[spd+1]
			, 1000000.0 / st_lamda.us_avg[spd+1]);
		write_log(buff, 1);
		buff.Printf(_T("    Middle  : %8d Cent:%6.1fHz Avg:%6.1fHz")
			, rep_p1_sample_num[2]
			, 1000000.0 / st_lamda.us_mid[spd]
			, 1000000.0 / st_lamda.us_mid_avg[spd]);
		write_log(buff, 1);
		buff.Printf(_T("    Too Long : %8d"),rep_p1_sample_num[3]);
		write_log(buff, 1);
		buff.Printf(_T("    Too Short: %8d"),rep_p1_sample_num[4]);
		write_log(buff, 1);
		buff.Printf(_T("    Error    : %8d"),rep_p1_sample_num[5]);
		write_log(buff, 1);

		write_log(_T(""), 1);
	}

	// phase2 report
	if (infile_type <= L3C && outfile_type >= WAV) {
		buff = _T(" [ l3c -> l3b, t9x ]");
		write_log(buff, 1);
//		buff.Printf(_T(" %d Baud"),(param.GetBaud() ? 1200 : 600));
//		write_log(buff, 1);

		if (carrier_data->total_r_pos > 0) {
			buff.Printf(_T(" %d / %d errors. (%.2f%%)"),rep_p2_error_num,carrier_data->total_r_pos,(rep_p2_error_num * 100.0 / carrier_data->total_r_pos));
			write_log(buff, 1);
		}

		write_log(_T(""), 1);
	}

	// phase3 report
	if (infile_type <= T9X && outfile_type >= WAV) {
		buff = _T(" [ l3b, t9x -> l3 ]");
		write_log(buff, 1);

		if ((param.GetWordSelect() & 0x04) == 0) {
			// 7bit
			buff = _T(" 7bits");
		} else {
			// 8bit
			buff = _T(" 8bits");
		}

		if (param.GetWordSelect() == 4 || param.GetWordSelect() == 5) {
			buff += _T(" NoParity");
		} else if ((param.GetWordSelect() & 0x01) == 0) {
			buff += _T(" EvenParity");
		} else {
			buff += _T(" OddParity");
		}
		 if ((param.GetWordSelect() & 0x03) == 0) {
			buff += _T(" 2stopbit");
		} else {
			buff += _T(" 1stopbit");
		}
		write_log(buff, 1);

		if (serial_data->total_r_pos > 0) {
			buff.Printf(_T(" %d / %d errors. (%.2f%%)"),rep_p3_error_num,serial_data->total_r_pos,(rep_p3_error_num * 100.0 / serial_data->total_r_pos));
			write_log(buff, 1);
		}

		write_log(_T(""), 1);
	}

	// phase4 report
	if (infile_type <= L3 && outfile_type >= WAV) {
		buff = _T(" [ l3 -> real data ]");
		write_log(buff, 1);

		if (rep4.empty()) {
			buff = _T(" data cannot parse.");
			write_log(buff, 1);
		} else {
			int idx = 1;
			std::vector<REPORT4 *>::iterator itm = rep4.begin();
			while ( itm != rep4.end() ) {
				buff.Printf(_T("%03d:"), idx);
				write_log(buff, 0);
				if ((*itm)->flags & 1) {
					buff = _T(" dataname: \"");
					buff += UTILS::conv_internal_name((*itm)->save_data_name);
					buff += _T("\"   [");
					switch((*itm)->save_data_name[8]) {
						case 0:
							buff += _T("BASIC");
							break;
						case 1:
							buff += _T("DATA");
							break;
						case 2:
							buff += _T("Machine");
							break;
						default:
							buff += _T("?");
							break;
					}
					switch((*itm)->save_data_name[9]) {
						case 0:
							buff += _T(" - Binary save");
							break;
						case 0xff:
							buff += _T(" - Ascii save");
							break;
						default:
							buff += _T(" ?");
							break;
					}
					buff += _T("]");
					write_log(buff, 0);

					int idx = ((*itm)->baud)-1;
					if (idx >= 0 && idx < 4) {
						buff.Printf(_T(" (%4dbaud)"), (int)c_baud_rate[idx] * (spd + 1));
					} else {
						buff = _T(" (----baud)");
					}
					write_log(buff, 1);
				} else {
					buff = _T(" no filename");
					write_log(buff, 1);
					buff = _T(" header section not found.");
					write_log(buff, 1);
				}
				if (((*itm)->flags & 2) == 0) {
					buff = _T(" data section not found.");
					write_log(buff, 1);
				}
				if (((*itm)->flags & 4) == 0) {
					buff = _T(" footer section not found.");
					write_log(buff, 1);
				}

				if ((*itm)->chksum_err_num <= 0) {
					buff = _T(" check sum ok.");
				} else {
					buff = _T(" check sum error exists.");
				}
				write_log(buff, 1);
				for(int i=0; i<(*itm)->chksum_err_num; i++) {
					buff.Printf(_T(" pos: %d-%d (%06x-%06x) ")
						,((*itm)->chksum_err_pos)[i].start_pos, ((*itm)->chksum_err_pos)[i].end_pos
						,((*itm)->chksum_err_pos)[i].start_pos, ((*itm)->chksum_err_pos)[i].end_pos);
					buff += UTILS::get_time_str(((*itm)->chksum_err_pos)[i].start_ms);
					buff += _T("-");
					buff += UTILS::get_time_str(((*itm)->chksum_err_pos)[i].end_ms);
					write_log(buff, 1);
				}

				itm++;
				idx++;
			}
		}

		write_log(_T(""), 1);
	}

	// phase3 report
	if (infile_type >= L3 && outfile_type <= T9X) {
		if (infile_type == PLAIN) {
			buff = _T(" [ plain");
		} else {
			buff = _T(" [ l3");
		}
		buff += _T(" -> l3b, t9x ]");
		write_log(buff, 1);

		if ((param.GetWordSelect() & 0x04) == 0) {
			// 7bit
			buff = _T(" 7bits");
		} else {
			// 8bit
			buff = _T(" 8bits");
		}

		if (param.GetWordSelect() == 4 || param.GetWordSelect() == 5) {
			buff += _T(" NoParity");
		} else if ((param.GetWordSelect() & 0x01) == 0) {
			buff += _T(" EvenParity");
		} else {
			buff += _T(" OddParity");
		}
		 if ((param.GetWordSelect() & 0x03) == 0) {
			buff += _T(" 2stopbit");
		} else {
			buff += _T(" 1stopbit");
		}
		write_log(buff, 1);

		write_log(_T(""), 1);
	}

	// phase2 report
	if (infile_type >= L3B && outfile_type <= L3C) {
		buff = _T(" [ l3b, t9x -> l3c ]");
		write_log(buff, 1);
		buff.Printf(_T(" %4d Baud"),(int)c_baud_rate[param.GetBaud()] * (spd + 1));
		write_log(buff, 1);

		write_log(_T(""), 1);
	}

	// phase1 report
	if (infile_type >= L3C && outfile_type <= WAV) {
		buff = _T(" [ l3c -> wav ]");
		write_log(buff, 1);
		buff.Printf(_T(" %dHz %dbit %dch"),outwav.GetSampleRate(), outwav.GetSampleBits(), outwav.GetChannels());
		write_log(buff, 1);

		write_log(_T(""), 1);
	}
}


/// @brief 波形解析レポートログ出力
///
///
///
///
void PARSEWAV::reporting_analyze()
{
	buff = _T("----- Result Report -----");
	write_log(buff, 1);

	// input
	buff = _T(" [ input ]");
	write_log(buff, 1);
	buff = _T(" ") + infile;
	write_log(buff, 1);
	if (infile_type == WAV) {
		buff.Printf(_T(" %dHz %dbit %dch"), inwav.GetSampleRate(), inwav.GetSampleBits(), inwav.GetChannels());
		write_log(buff, 1);
	}

	for(int spd=0; spd<2; spd++) {
		write_log(_T(""), 1);
		buff = _T("  Wave Type: ");
		buff += spd ? _T("Double Speed FSK") : _T("Standard FSK");
		write_log(buff, 1);
		for(int i=0; i<2; i++) {
			switch(i) {
				case 0:
					buff = _T("    Cos Wave:");
					break;
				case 1:
					buff = _T("    Sin Wave:");
					break;
			}
			write_log(buff, 1);
			buff.Printf(_T("      Long(0)  : %8d Cent:%6.1fHz Avg:%6.1fHz")
				, st_chkwav[spd].sample_num[i][0]
				, 1000000.0 / st_lamda.us[spd]
				, 1000000.0 / st_chkwav[spd].us0avg[i]);
			write_log(buff, 1);
			buff.Printf(_T("      Short(1) : %8d Cent:%6.1fHz Avg:%6.1fHz")
				, st_chkwav[spd].sample_num[i][1]
				, 1000000.0 / st_lamda.us[spd+1]
				, 1000000.0 / st_chkwav[spd].us1avg[i]);
			write_log(buff, 1);
			buff.Printf(_T("      Middle   : %8d"),st_chkwav[spd].sample_num[i][2]);
			write_log(buff, 1);
			buff.Printf(_T("      Too Long : %8d"),st_chkwav[spd].sample_num[i][3]);
			write_log(buff, 1);
			buff.Printf(_T("      Too Short: %8d"),st_chkwav[spd].sample_num[i][4]);
			write_log(buff, 1);
			buff.Printf(_T("      %dbaud: %4d  %dbaud: %4d  %dbaud: %4d  %dbaud: %4d")
				,(int)c_baud_rate[0] * (spd + 1)
				,st_chkwav[spd].baud_num[i][0]
				,(int)c_baud_rate[1] * (spd + 1)
				,st_chkwav[spd].baud_num[i][1]
				,(int)c_baud_rate[2] * (spd + 1)
				,st_chkwav[spd].baud_num[i][2]
				,(int)c_baud_rate[3] * (spd + 1)
				,st_chkwav[spd].baud_num[i][3]
			);
			write_log(buff, 1);
			buff.Printf(_T("      Normal : %4d  Reverse : %4d  AmpMax : %4d  AmpMin : %4d")
				,st_chkwav[spd].rev_num[i][0], st_chkwav[spd].rev_num[i][1]
				,st_chkwav[spd].amp_max[i], st_chkwav[spd].amp_min[i]
			);
			write_log(buff, 1);
		}
	}
	write_log(_T(""), 1);

	// report
	if (infile_type <= WAV && outfile_type >= WAV) {
		int fsk_spd = param.GetFskSpeed();

		buff = _T(" [ result ]");
		write_log(buff, 1);

		buff = _T("  Wave Type: ");
		buff += fsk_spd ? _T("Double Speed FSK") : _T("Standard FSK");
		write_log(buff, 1);

		buff.Printf(_T("  Wave Reverse: %s"),(param.GetReverseWave() ? _T("on") : _T("off")));
		write_log(buff, 1);
		buff.Printf(_T("  Correct Type: %s"),(param.GetCorrectType() ? _T("sin wave") : _T("cos wave")));
		write_log(buff, 1);
//		buff.Printf(_T("  Avg: %4dHz  %4dHz")
//			,param.freq1200,param.freq2400);
//		write_log(buff, 1);

		buff.Printf(_T("  %4d Baud"),(int)c_baud_rate[param.GetBaud()] * (fsk_spd + 1));
		write_log(buff, 1);

		write_log(_T(""), 1);
	}
}

/// @brief ログファイルに出力
///
/// @param[in] buff メッセージ
/// @param[in] crlf 改行する行数
///
void PARSEWAV::write_log(const wxString &buff, int crlf)
{
	if (buff.Length() > 0) {
		if (fiolog) _fputts(buff.c_str(), fiolog);
		if (logbuf) *logbuf += buff;
	}
	for (int i=0; i<crlf; i++) {
		if (fiolog) _fputts(_T("\n"), fiolog);
#ifdef _WIN32
		if (logbuf) *logbuf += _T("\r\n");
#else
		if (logbuf) *logbuf += _T("\n");
#endif
	}
}

/// @brief ファイル名取得
///
/// @param[out] filename ファイル名
///
///
void PARSEWAV::GetFileNameBase(wxString &filename)
{
	UTILS::prefix_name(infile, filename);
}

/// @brief 拡張子のチェック
///
/// @param[in] filename ファイル名
/// @param[in] ext 拡張子名
/// @return true 拡張子が一致 / false 一致しない
bool PARSEWAV::check_extension(const wxString &filename, const wxString &ext)
{
	size_t nam_len = filename.Length();
	size_t ext_len = ext.Length();

	if(nam_len >= ext_len) {
		if (filename.Right(ext_len).IsSameAs(ext, false) == true) {
			return true;
		}
	}
	return false;
}

/// @brief ログバッファへのポインタを設定
///
/// @param[in,out] buf バッファポインタ
///
///
void PARSEWAV::SetLogBufferPtr(wxString *buf)
{
	logbuf = buf;
}

/// @brief 実ファイルのファイルフォーマット種類を得る
///
/// @return 種類 0 1 2
///
///
int  PARSEWAV::GetRfDataFormat()
{
	return rftypebox->getRfDataFormat();
}

/// @brief 実ファイルのファイルフォーマット種類ダイアログを表示
///
/// @return true OK / false キャンセル
///
///
bool PARSEWAV::ShowRfTypeBox()
{
	// ファイルの種類を選択
	int rc = rftypebox->showRftypeBox((_TCHAR *)NULL, false);
	if (rc != 1) {
		// cancel button
		return false;
	}

	if (fio_in != NULL) {
		PwErrType rt = get_first_rf_data(fio_in, &infile_type);
		if (rt == pwError) {
			return false;
		}
	}
	return true;
}

/// @brief マシン語情報ダイアログを表示
///
/// @param[in] hide_no_header_info ヘッダなし情報を表示しないか
/// @return true OK / false キャンセル
///
bool PARSEWAV::ShowMAddressBox(bool hide_no_header_info) {
	int rc;
	int addr;

	rc = maddressbox->showMAddressBox(hide_no_header_info);
	if (rc == 1) {
		rf_header[0] = 0;
		addr = (int)maddressbox->getStartAddr();
		rf_header[3] = addr / 256;
		rf_header[4] = addr % 256;
		addr = (int)maddressbox->getDataSize();
		rf_header[1] = addr / 256;
		rf_header[2] = addr % 256;

		rf_footer[0] = 0xff;
		rf_footer[1] = 0;
		rf_footer[2] = 0;
		addr = (int)maddressbox->getExecAddr();
		rf_footer[3] = addr / 256;
		rf_footer[4] = addr % 256;

		include_header = maddressbox->getIncHeader();

		return true;
	}
	return false;
}

/// @brief 音データからバイナリデータに変換
///
/// @return pwOK 正常
///
///
PwErrType PARSEWAV::DecodeData()
{
	PwErrType rc = pwOK;

	wave_data->init();
	wave_correct_data->init();
	carrier_data->init();
	serial_data->init();
	serial_new_data->init();
	binary_data->init();

	phase2_frip = 0;

	phase3_baud = 0;
	phase2n_baud_count = 0;

	half_wave = param.GetHalfWave();

	memset(rep_p1_sample_num, 0, sizeof(rep_p1_sample_num));
	rep_p2_error_num = 0;
	rep_p3_error_num = 0;

	std::vector<REPORT4 *>::iterator itm = rep4.begin();
	while ( itm != rep4.end() ) {
		delete *itm;
		itm++;
	}
	rep4.clear();
	rep4_itm = NULL;

	memset(save_data_name, 0, sizeof(save_data_name));
	save_data_name[8]=0x7f;
	save_data_name[9]=0x7f;

	memset(&inwav, 0, sizeof(inwav));

	// 入力ファイル先頭にセット
	fseek(fio_in, 0, SEEK_SET);
	sample_pos = 0;
	write_pos = 0;
	sample_ms = 0;

	// 入力ファイルのシーク
	if ((rc = SeekFileFormat(fio_in, infile, infile_type)) != pwOK) {
		return rc;
	}

	// ファイルのヘッダを出力
	if (outfile_type >= infile_type) {
		InitFileHeader(fio_out, outfile_type);
	}

	progbox->initProgress(outfile_type == NO_FILE ? 1 : 0, 0, 100);

	int fsk_spd = param.GetFskSpeed();

	if (infile_type == WAV) {
		// wav
		phase1 = 0;
		phase2 = 0;
		phase2n = 0;
		phase3 = 0;
		phase4 = 0;
		decode_phase1(fsk_spd, wave_data, wave_correct_data, carrier_data, serial_data, serial_new_data, binary_data, 0);
	} else if (infile_type == L3C) {
		// l3c
		phase1 = 11;
		phase2 = 0;
		phase2n = 0;
		phase3 = 0;
		phase4 = 0;
		decode_phase1(fsk_spd, wave_data, wave_correct_data, carrier_data, serial_data, serial_new_data, binary_data, 11);
	} else if (infile_type == L3B) {
		// l3b
		phase1 = 11;
		phase2 = 21;
		phase2n = 0;
		phase3 = 0;
		phase4 = 0;
		carrier_data->last = true;
		decode_phase2(fsk_spd, carrier_data, serial_data, serial_new_data, binary_data, 21);
	} else if (infile_type == T9X) {
		// t9x
		phase1 = 11;
		phase2 = 22;
		phase2n = 0;
		phase3 = 0;
		phase4 = 0;
		carrier_data->last = true;
		decode_phase2(fsk_spd, carrier_data, serial_data, serial_new_data, binary_data, 22);
	} else if (infile_type == L3) {
		// l3
		phase1 = 11;
		phase2 = 21;
		phase2n = 0;
		phase3 = 31;
		phase4 = 0;
		serial_data->last = true;
		decode_phase3(serial_data, binary_data, 31);
	}

	// ファイルのヘッダを更新
	if (outfile_type >= infile_type) {
		SetFileHeader(fio_out, infile_type, outfile_type);
	}
	return rc;
}

/// @brief バイナリデータから音データに変換
///
/// @return pwOK:OK pwCancel:キャンセルボタンを押した
///
///
PwErrType PARSEWAV::EncodeData()
{
	PwErrType rc = pwOK;

	wave_data->init();
	carrier_data->init();
	serial_data->init();
	serial_new_data->init();
	binary_data->init();

	phase2_frip = 0;
	phase3_baud = 0;
	phase2n_baud_count = 0;
	baud24_frip = 0;

	memset(rep_p1_sample_num, 0, sizeof(rep_p1_sample_num));
	rep_p2_error_num = 0;
	rep_p3_error_num = 0;

	// 実ファイルの情報をセット
	memset(save_data_name, 0, sizeof(save_data_name));
	if (infile_type == REAL) {
		rftypebox->getRfDataName(save_data_name);
		save_data_name[8]=rftypebox->getRfDataFormat();
		save_data_name[9]=rftypebox->getRfDataType();
		save_data_name[10]=rftypebox->getRfDataType();

		// マシン語の場合
		if (save_data_name[8] == 2 && maddressbox->getValid() != true) {
			// ヘッダがない場合、情報入力を促す
			if (ShowMAddressBox(false) == false) {
				include_header = true;
//				return pwCancel;
			}
		}

		if (include_header) {
			sample_num = sample_num_orig;
		} else {
			sample_num = sample_num_orig + 10;
		}
	}

	fseek(fio_in, 0, SEEK_SET);
	sample_pos = 0;
	write_pos = 0;
	sample_ms = 0;

	// 入力ファイルの位置をシーク
	if ((rc = SeekFileFormat(fio_in, infile, infile_type)) != pwOK) {
		return rc;
	}

	// ファイルのヘッダを出力
	InitFileHeader(fio_out, outfile_type);

	progbox->initProgress(outfile_type == NO_FILE ? 1 : 0, 0, 100);

	if (infile_type == L3C) {
		// l3c
		phase4 = 0;
		phase3 = 20;
		phase2n = 0;
		phase2 = 10;
		phase1 = 11;
		encode_phase2(serial_data, carrier_data, 10, true);
	} else if (infile_type == T9X) {
		// t9x
		phase4 = 0;
		phase3 = 21;
		phase2n = 0;
		phase2 = 21;
		phase1 = 11;
		encode_phase3(binary_data, serial_data, carrier_data, 21);
	} else if (infile_type == L3B) {
		// l3b
		phase4 = 0;
		phase3 = 20;
		phase2n = 0;
		phase2 = 21;
		phase1 = 11;
		encode_phase3(binary_data, serial_data, carrier_data, 20);
	} else if (infile_type == L3) {
		// l3
		phase4 = 0;
		phase3 = 30;
		phase2n = 0;
		phase2 = 21;
		phase1 = 11;
		encode_phase3(binary_data, serial_data, carrier_data, 30);
	} else if (infile_type == PLAIN) {
		// plain
		phase4 = 0;
		phase3 = 30;
		phase2n = 0;
		phase2 = 21;
		phase1 = 11;
		encode_phase3(binary_data, serial_data, carrier_data, 30);
	} else if (infile_type == REAL) {
		// real
		phase4 = 0;
		phase3 = 40;
		phase2n = 0;
		phase2 = 21;
		phase1 = 11;
		encode_phase4(binary_data, serial_data, carrier_data);
	}

	// 必要ならファイルのヘッダを更新
	SetFileHeader(fio_out, infile_type, outfile_type);

	return pwOK;
}

/// @brief 波形を解析
///
/// @return 0
///
///
int PARSEWAV::AnalyzeWave()
{
	int rc = 0;
	int correct_type = 0;
	int reverse = 0;

	chk_wav_mode = true;

	outfile_type = NO_FILE;	// 出力ファイルなし

	std::vector<REPORT4 *>::iterator itm = rep4.begin();
	while ( itm != rep4.end() ) {
		delete *itm;
		itm++;
	}
	rep4.clear();
	rep4_itm = NULL;

	half_wave = 1;

	memset(save_data_name, 0, sizeof(save_data_name));
	save_data_name[8]=0x7f;
	save_data_name[9]=0x7f;

	if (logbuf) {
		*logbuf = _T("");
	}

	memset(st_chkwav, 0, sizeof(st_chkwav));

	progbox->initProgress(1, 0, 100);

	for(int spd=0; spd < 2; spd++) {
		param.SetFskSpeed(spd);
		for(int i=0; i < 2; i++) {
			st_chkwav[spd].num = i;
			// fseekする
			fseek(fio_in, 0, SEEK_SET);
			sample_pos = 0;
			write_pos = 0;
			sample_ms = 0;

			check_wav_format(fio_in, inwav.GetHead(), inwav.GetFmtChank(), inwav.GetDataChank());

			wave_data->init();
			wave_correct_data->init();
			carrier_data->init();
			serial_data->init();
			serial_new_data->init();
			binary_data->init();

			st_chkwav[spd].analyze_num = inwav.GetSampleRate() * 30;	// 30秒解析

			phase2_frip = 0;
			phase3_baud = 0;
			phase2n_baud_count = 0;

			memset(rep_p1_sample_num, 0, sizeof(rep_p1_sample_num));
			rep_p2_error_num = 0;
			rep_p3_error_num = 0;

			correct_type = i % 2;
			reverse = 0;

			phase1 = 0;
			phase2 = 11;
			phase2n = 0;
			phase3 = 0;
			phase4 = 0;

			decode_phase1(spd, wave_data, wave_correct_data, carrier_data, serial_data, serial_new_data, binary_data, 0, correct_type, reverse);
			st_chkwav[spd].sample_num[i][0]=rep_p1_sample_num[0];
			st_chkwav[spd].sample_num[i][1]=rep_p1_sample_num[1];
			st_chkwav[spd].sample_num[i][2]=rep_p1_sample_num[2];
			st_chkwav[spd].sample_num[i][3]=rep_p1_sample_num[3];
			st_chkwav[spd].sample_num[i][4]=rep_p1_sample_num[4];
			st_chkwav[spd].us0avg[i] = st_lamda.us_avg[spd];
			st_chkwav[spd].us1avg[i] = st_lamda.us_avg[spd+1];
			st_chkwav[spd].amp_max[i] = dft.GetAmpMax();
			st_chkwav[spd].amp_min[i] = dft.GetAmpMin();
		}
	}

	progbox->endProgress();

	// 尤もらしいものを判定
	struct st_judge_params {
		int best;
		int spd;
		int cor;
	};
	// 0: best baud
	// 1: max amp
	// 2: min amp
	// 3: long sample
	// 4: short sample
	struct st_judge_params best[5];
	for(int i=0; i<5; i++) {
		best[i].best = 0;
		best[i].spd = -1;
		best[i].cor = -1;
	}
	int best_spd_cnt[2] = {0,0};
	int best_cor_cnt[2] = {0,0};
	for(int spd=0; spd<2; spd++) {
		for(int cor=0; cor<2; cor++) {
			int baud_num = 0;
			for(int b=0; b<4; b++) {
				baud_num += st_chkwav[spd].baud_num[cor][b];
			}
			if (best[0].best < baud_num) {
				best[0].best = baud_num;
				best[0].spd = spd;
				best[0].cor = cor;
			}
			if (best[1].best < st_chkwav[spd].amp_max[cor]) {
				best[1].best = st_chkwav[spd].amp_max[cor];
				best[1].spd = spd;
				best[1].cor = cor;
			}
			if (best[2].best > st_chkwav[spd].amp_min[cor]) {
				best[2].best = st_chkwav[spd].amp_min[cor];
				best[2].spd = spd;
				best[2].cor = cor;
			}
#if 0
			if (best[3].best < (st_chkwav[spd].sample_num[cor][0] << (1-spd))) {
				best[3].best = st_chkwav[spd].sample_num[cor][0] << (1-spd);
				best[3].spd = spd;
				best[3].cor = cor;
			}
			if (best[4].best < (st_chkwav[spd].sample_num[cor][1] << (1-spd))) {
				best[4].best = (st_chkwav[spd].sample_num[cor][1] << (1-spd));
				best[4].spd = spd;
				best[4].cor = cor;
			}
#endif
		}
	}
	for(int i=0; i<5; i++) {
		if (best[i].spd >= 0) best_spd_cnt[best[i].spd]++;
		if (best[i].cor >= 0) best_cor_cnt[best[i].cor]++;
	}

	int best_fsk_spd = best_spd_cnt[0] < best_spd_cnt[1] ? 1 : 0;
	int best_correct = best_cor_cnt[0] < best_cor_cnt[1] ? 1 : 0;

	param.SetFskSpeed(best_fsk_spd);

#if 0
	// 一番適切な組み合わせを探す
	long long long_data = 0;
	long long mid_data = sample_num;
	int long_num = -1;
	int mid_num = 0;
	int best_num = 0;
	// sin or cos ?
	for(int i=1; i >= 0; i--) {
		// long波の個数
		if (long_data < (st_chkwav[best_fsk_spd].sample_num[i][0] / 10)) {
			// 最大値
			long_data = (st_chkwav[best_fsk_spd].sample_num[i][0] / 10);
			long_num = i;
		}
		if (mid_data >= (st_chkwav[best_fsk_spd].sample_num[i][2] / 10)) {
			// middle波の個数
			mid_data = (st_chkwav[best_fsk_spd].sample_num[i][2] / 10);
			mid_num = i;
		} else if (mid_data == (st_chkwav[best_fsk_spd].sample_num[i][2] / 10)) {
			// 同じ値がある場合
			mid_num = -1;
		}
	}
	if (mid_num >= 0) {
		best_num = mid_num;
	} else {
		best_num = long_num;
	}
#endif

	// ボーレート判定
	int best_baud = -1;
	int baud_cnt = 0;
	for(int i=0; i<4; i++) {
		int idx = c_baud_min[i];
		int num = st_chkwav[best_fsk_spd].baud_num[best_correct][idx];
		if (num >= baud_cnt) {
			best_baud = idx;
			baud_cnt = num;
		}
	}
	if (best_baud >= 0) {
		param.SetBaud(best_baud);
	}

	// 波正逆判定
	if (st_chkwav[best_fsk_spd].rev_num[best_correct][0] < st_chkwav[best_fsk_spd].rev_num[best_correct][1]) {
		// 逆
		param.SetReverseWave(1);
	} else {
		// 正
		param.SetReverseWave(0);
	}

	// パラメータをセット
	param.SetCorrectWave(1);
	param.SetCorrectType(best_correct % 2);

	// 波長の平均値を設定
//	param.freq1200 = (int)(1000000.0 / st_chkwav.us1200avg[best_num] /  (1 << half_wave));
//	param.freq2400 = (int)(1000000.0 / st_chkwav.us2400avg[best_num] /  (1 << half_wave));

	// レポート
	reporting_analyze();

	chk_wav_mode = false;

	return rc;
}

/// @brief 変換したデータをファイルに出力
///
/// @return true 正常 / false エラーあり
///
///
bool PARSEWAV::ExportData(enum_file_type file_type)
{
	PwErrType rc = pwOK;
	if (file_type != UNKNOWN) {
		outfile_type = file_type;
	}

	// デバッグログオープン
	if (param.GetDebugMode()) {
		if (!(fiolog = UTILS::pw_fopen(logfile, _T("w")))) {
			fiolog = NULL;
			err_num = pwErrCannotWriteDebugLog;
			errinfo->SetInfo(__LINE__, pwWarning, err_num);
			errinfo->ShowMsgBox();
		}
	}

	if (logbuf) {
		*logbuf = _T("");
	}

	if (infile_type <= outfile_type) {
		// decode
		rc = DecodeData();
	} else if (infile_type > outfile_type) {
		// encode
		rc = EncodeData();
		// decode チェック
		if (rc == pwOK) rc = DecodeData();
	}

	reporting();

	if (fiolog != NULL) {
		fclose(fiolog);
		fiolog = NULL;
	}

	progbox->endProgress();

	if (rc == pwCancel || rc == pwError) {
		return false;
	}

	return true;
}

/// @brief 出力用ファイルを開く
///
/// @param[in] out_file ファイルパス名
/// @return true:正常 false:エラーあり
///
bool PARSEWAV::OpenOutFile(wxString &out_file)
{
	// 同じファイルはダメ
	wxFileName infile_name = wxFileName::FileName(infile);
	wxFileName outfile_name = wxFileName::FileName(out_file);
	if (outfile_name.SameAs(infile_name)) {
		err_num = pwErrSameFile;
		errinfo->SetInfo(__LINE__, pwError, err_num);
		errinfo->ShowMsgBox();
		return false;
	}

	logfile = out_file + _T(".log");

	outfile_type = UNKNOWN;
	if (check_extension(out_file, _T(".WAV"))) {
		outfile_type = WAV;
	} else if (check_extension(out_file, _T(".L3C"))) {
		outfile_type = L3C;
	} else if (check_extension(out_file, _T(".L3B"))) {
		outfile_type = L3B;
	} else if (check_extension(out_file, _T(".T9X"))) {
		outfile_type = T9X;
	} else if (check_extension(out_file, _T(".L3"))) {
		outfile_type = L3;
	} else {
		outfile_type = REAL;
	}

	fio_out  = NULL;
	fio_outs = NULL;
	if (param.GetFileSplit() && outfile_type == REAL) {
		// 分割する場合
#ifdef _WIN32
		int seppos = out_file.Find(wxChar('\\'), true);
#else
		int seppos = out_file.Find(wxChar('/'), true);
#endif
		int pos = out_file.Find(wxChar('.'), true);
		if (pos != wxNOT_FOUND && seppos < pos) {
			outsext = out_file.Mid(pos);
			outsfileb = out_file.Left(pos);
		} else {
			outsext = _T("");
			outsfileb = out_file;
		}
		outsfile.Printf(_T("_%03d"), 1);
		outsfile = outsfileb + outsfile;
		outsfile += outsext;
		if (!(fio_outs = UTILS::pw_fopen(outsfile, _T("wb")))) {
			err_num = pwErrCannotWrite;
			errinfo->SetInfo(__LINE__, pwError, err_num);
			errinfo->ShowMsgBox();
			return false;
		}
		fclose(fio_outs);
		fio_outs = NULL;

	} else {
		// 分割しない
		if (!(fio_out = UTILS::pw_fopen(out_file, _T("wb")))) {
			err_num = pwErrCannotWrite;
			errinfo->SetInfo(__LINE__, pwError, err_num);
			errinfo->ShowMsgBox();
			return false;
		}
	}

	outfile = out_file;

	return true;
}

/// @brief 出力用ファイルを閉じる
///
///
///
///
void PARSEWAV::CloseOutFile()
{
	if (fio_outs != NULL) {
		fclose(fio_outs);
		fio_outs = NULL;
	}
	if (fio_out != NULL) {
		fclose(fio_out);
		fio_out = NULL;
	}
	logfile = _T("wavtool.log");
}

/// @brief 入力ファイルを開く
///
/// @param[in] in_file ファイルパス名
/// @return true:正常 false:エラー
///
bool PARSEWAV::OpenDataFile(wxString &in_file)
{
	PwErrType rc;


	infile_type = UNKNOWN;
	if (check_extension(in_file, _T(".WAV"))) {
		infile_type = WAV;
	} else if (check_extension(in_file, _T(".L3C"))) {
		infile_type = L3C;
	} else if (check_extension(in_file, _T(".L3B"))) {
		infile_type = L3B;
	} else if (check_extension(in_file, _T(".T9X"))) {
		infile_type = T9X;
	} else if (check_extension(in_file, _T(".L3"))) {
		infile_type = L3;
	} else {
		infile_type = REAL;
	}

	CloseDataFile();

	if (!(fio_in = UTILS::pw_fopen(in_file, _T("rb")))) {
		err_num = pwErrFileNotFound;
		errinfo->SetInfo(__LINE__, pwError, err_num);
		errinfo->ShowMsgBox();
		return false;
	}

	// ファイルのチェック
	rc = CheckFileFormat(fio_in, in_file, &infile_type);
	if (rc == pwError || rc == pwCancel) {
		CloseDataFile();
		return false;
	}

	infile = in_file;

//	param.freq1200 = 1200;
//	param.freq2400 = 2400;

	return true;

}

/// @brief 入力ファイルを閉じる
///
///
///
///
void PARSEWAV::CloseDataFile()
{
	if (fio_in != NULL) {
		fclose(fio_in);
		fio_in = NULL;
	}
}

